from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import range
from builtins import object
import gensim
import copy
import pandas as pd

from corextopic import corextopic
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import LatentDirichletAllocation, NMF
from collections import defaultdict

from pewtils import is_not_null


class TopicModel(object):

    """
    A wrapper around various topic modeling algorithms and libraries, intended to provide a standardized way to train \
    and apply models. When you initialize a ``TopicModel``, it will fit a vectorizer, and split the data into a train \
    and test set if ``holdout_pct`` is provided. For more information about the available implementations, refer to the \
    documentation for the ``fit()`` method below.

    :param df: A :py:class:`pandas.DataFrame`
    :param text_col: Name of the column containing text
    :type text_col: str
    :param method: The topic model implementation to use. Options are: sklearn_lda, sklearn_nmf, gensim_lda, \
    gensim_hdp, corex
    :type method: str
    :param num_topics: The number of topics to extract. Required for every method except ``gensim_hdp``.
    :type num_topics: int
    :param max_ngram_size: Maximum ngram size (2=bigrams, 3=trigrams, etc)
    :type max_ngram_size: int
    :param holdout_pct: Proportion of the documents to hold out for goodness-of-fit scoring
    :type holdout_pct: float
    :param use_tfidf: Whether to use binary counts or a TF-IDF representation
    :type use_tfidf: bool
    :param vec_kwargs: All remaining arguments get passed to TfidfVectorizer or CountVectorizer

    Usage::

        from pewanalytics.text.topics import TopicModel

        import nltk
        import pandas as pd
        nltk.download("movie_reviews")
        reviews = [{"fileid": fileid, "text": nltk.corpus.movie_reviews.raw(fileid)} for fileid in nltk.corpus.movie_reviews.fileids()]
        df = pd.DataFrame(reviews)

        >>> model = TopicModel(df, "text", "sklearn_nmf", num_topics=5, min_df=25, max_df=.5, use_tfidf=False)
        Initialized sklearn_nmf topic model with 3285 features
        1600 training documents, 400 testing documents

        >>> model.fit()

        >>> model.print_topics()
        0: bad, really, know, don, plot, people, scene, movies, action, scenes
        1: star, trek, star trek, effects, wars, star wars, special, special effects, movies, series
        2: jackie, films, chan, jackie chan, hong, master, drunken, action, tarantino, brown
        3: life, man, best, characters, new, love, world, little, does, great
        4: alien, series, aliens, characters, films, television, files, quite, mars, action

        >>> doc_topics = model.get_document_topics(df)

        >>> doc_topics
               topic_0   topic_1   topic_2   topic_3   topic_4
        0     0.723439  0.000000  0.000000  0.000000  0.000000
        1     0.289801  0.050055  0.000000  0.000000  0.000000
        2     0.375149  0.000000  0.030691  0.059088  0.143679
        3     0.152961  0.010386  0.000000  0.121412  0.015865
        4     0.294005  0.100426  0.000000  0.137630  0.051241
        ...        ...       ...       ...       ...       ...
        1995  0.480983  0.070431  0.135178  0.256951  0.000000
        1996  0.139986  0.000000  0.000000  0.107430  0.000000
        1997  0.141545  0.005990  0.081986  0.387859  0.057025
        1998  0.029228  0.023342  0.043713  0.280877  0.107551
        1999  0.044863  0.000000  0.000000  0.718677  0.000000

    """

    def __init__(
        self,
        df,
        text_col,
        method,
        num_topics=None,
        max_ngram_size=2,
        holdout_pct=0.25,
        use_tfidf=False,
        **vec_kwargs
    ):

        self.df = df
        self.text_col = text_col
        self.method = method
        self.num_topics = num_topics
        self.train_df = df.sample(int(round(len(df) * (1.0 - holdout_pct))))
        self.train_df = self.train_df.dropna(subset=[self.text_col])
        self.test_df = df[~df.index.isin(self.train_df.index)]
        self.test_df = self.test_df.dropna(subset=[self.text_col])
        if "stop_words" not in vec_kwargs:
            vec_kwargs["stop_words"] = "english"

        if use_tfidf:
            vec = TfidfVectorizer
        else:
            vec = CountVectorizer
        self.vectorizer = vec(
            ngram_range=(1, max_ngram_size), decode_error="ignore", **vec_kwargs
        )

        self.vectorizer = self.vectorizer.fit(self.train_df[self.text_col])
        self.ngrams = self.vectorizer.get_feature_names()
        if self.method in ["gensim_lda", "gensim_hdp"]:
            self.train_features = self.get_features(self.train_df, keep_sparse=True)
            self.test_features = self.get_features(self.test_df, keep_sparse=True)
            if self.method == "gensim_hdp":
                self.topic_ids = None
                if num_topics:
                    raise Exception(
                        "You cannot specify the number of topics for an HDP model"
                    )
        else:
            self.train_features = self.get_features(self.train_df)
            self.test_features = self.get_features(self.test_df)

        self.model = None

        print(
            "Initialized {} topic model with {} features".format(
                self.method, len(self.ngrams)
            )
        )
        try:
            print(
                "{} training documents, {} testing documents".format(
                    len(self.train_features), len(self.test_features)
                )
            )
        except TypeError:
            print(
                "{} training documents, {} testing documents".format(
                    self.train_features.shape[0], self.test_features.shape[0]
                )
            )

    def get_features(self, df, keep_sparse=False):

        """
        Uses the trained vectorizer to process a :py:class:`pandas.DataFrame` and return a feature matrix.

        :param df: The :py:class:`pandas.DataFrame` to vectorize (must have ``self.text_col`` in it)
        :param keep_sparse: Whether or not to keep the feature matrix in sparse format (default=False)
        :type keep_sparse: bool
        :return: A :py:class:`pandas.DataFrame` of features or a sparse matrix, depending on the value of \
        ``keep_sparse``
        """

        subset_df = df.dropna(subset=[self.text_col])
        features = self.vectorizer.transform(subset_df[self.text_col])
        if keep_sparse:
            return features
        else:
            return pd.DataFrame(
                features.todense(), columns=self.ngrams, index=subset_df.index
            )

    def get_fit_params(self, **kwargs):

        """
        Internal helper function to set defaults depending on the specified model.

        :param kwargs: Arguments passed to ``self.fit()``
        :return: Arguments to pass to the model
        """

        defaults = {
            "sklearn_lda": {
                "alpha": 1.0,
                "beta": 1.0,
                "learning_decay": 0.7,
                "learning_offset": 50,
                "learning_method": "online",
                "max_iter": 500,
                "batch_size": 128,
                "verbose": False,
            },
            "sklearn_nmf": {
                "alpha": 0.0,
                "l1_ratio": 0.5,
                "tol": 0.00001,
                "max_iter": 500,
                "shuffle": True,
            },
            "gensim_lda": {
                "chunksize": 1000,
                "passes": 10,
                "decay": 0.8,
                "offset": 1,
                "workers": 2,
                "alpha": None,
                "beta": "auto",
                "use_multicore": False,
            },
            "gensim_hdp": {
                "max_chunks": None,
                "max_time": None,
                "chunksize": 256,
                "kappa": 1.0,
                "tau": 64.0,
                "T": 150,
                "K": 15,
                "alpha": 1,
                "beta": 0.01,
                "gamma": 1,
                "scale": 1.0,
                "var_converge": 0.0001,
            },
            "corex": {"anchors": [], "anchor_strength": 3},
        }

        for k, v in kwargs.items():
            if k not in defaults[self.method].keys():
                raise Exception(
                    "Unknown keyword argument for method '{}': {}. Accepted parameters are: {}".format(
                        self.method, k, defaults[self.method].keys()
                    )
                )
        fit_params = copy.deepcopy(defaults[self.method])
        fit_params.update(kwargs)

        if self.method == "sklearn_lda":
            fit_params["verbose"] = int(fit_params["verbose"])
            if "alpha" in fit_params.keys():
                fit_params["doc_topic_prior"] = fit_params["alpha"] / float(
                    self.num_topics
                )
                del fit_params["alpha"]
            if "beta" in fit_params.keys():
                fit_params["topic_word_prior"] = fit_params["beta"] / float(
                    self.num_topics
                )
                del fit_params["beta"]

        if self.method == "gensim_lda":
            if not fit_params["alpha"]:
                if fit_params["use_multicore"]:
                    fit_params["alpha"] = "symmetric"
                else:
                    fit_params["alpha"] = "auto"

        if self.method in ["gensim_lda", "gensim_hdp"]:
            if "beta" in fit_params.keys():
                fit_params["eta"] = fit_params["beta"]
                del fit_params["beta"]

        return fit_params

    def fit(self, df=None, **kwargs):

        """
        Fits a model using the method specified when initializing the ``TopicModel``. Details on model-specific \
        parameters are below:

        **sklearn_lda**

        Fits a model using :py:class:`sklearn.decomposition.LatentDirichletAllocation`. For more information on \
        available parameters, please refer to the official documentation: \
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html

        :param df: The :py:class:`pandas.DataFrame` to train the model on (must contain ``self.text_col``)
        :param alpha: Represents document-topic density. When values are higher, documents will be comprised of more \
        topics; when values are lower, documents will be primarily comprised of only a few topics. This parameter is \
        used instead of the doc_topic_prior sklearn parameter, and will be passed along to sklearn using the formula: \
        ``doc_topic_prior = alpha / num_topics``
        :param beta: Represents topic-word density. When values are higher, topics will be comprised of more words; \
        when values are lower, only a few words will be loaded onto each topic. This parameter is used instead of the \
        topic_word_prior sklearn parameter, and will be passed along to sklearn using the formula: \
        ``topic_word_prior = beta / num_topics``.
        :param learning_decay: See sklearn documentation.
        :param learning_offset: See sklearn documentation.
        :param learning_method: See sklearn documentation.
        :param max_iter: See sklearn documentation.
        :param batch_size: See sklearn documentation.
        :param verbose: See sklearn documentation.

        **sklearn_nmf**

        Fits a model using :py:class:`sklearn.decomposition.NMF`. For more information on available parameters, \
        please refer to the official documentation: \
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html

        :param df: The :py:class:`pandas.DataFrame` to train the model on (must contain ``self.text_col``)
        :param alpha: See sklearn documentation.
        :param l1_ratio: See sklearn documentation.
        :param tol: See sklearn documentation.
        :param max_iter: See sklearn documentation.
        :param shuffle: See sklearn documentation.

        **gensim_lda**

        Fits an LDA model using :py:class:`gensim.models.LdaModel` or \
        :py:class:`gensim.models.ldamulticore.LdaMulticore`. \
        When ``use_multicore`` is set to True, the multicore implementation will be used, otherwise the standard \
        LDA implementation will be used. \
        For more information on available parameters, please refer to the official documentation below:

            - use_multicore=True: https://radimrehurek.com/gensim/models/ldamulticore.html
            - use_multicore=False: https://radimrehurek.com/gensim/models/ldamodel.html

        :param df: The :py:class:`pandas.DataFrame` to train the model on (must contain ``self.text_col``)
        :param alpha: Represents document-topic density. When values are higher, documents will be comprised of \
        more topics; when values are lower, documents will be primarily comprised of only a few topics. Gensim \
        options are a bit different than sklearn though; refer to the documentation for the accepted values here.
        :param beta: Represents topic-word density. When values are higher, topics will be comprised of more words; \
        when values are lower, only a few words will be loaded onto each topic. Gensim options are a bit different \
        than sklearn though; refer to the documentation for the accepted values here. Gensim calls this parameter \
        ``eta``. We renamed it to be consistent with the sklearn implementations.
        :param chunksize: See gensim documentation.
        :param passes: See gensim documentation.
        :param decay: See gensim documentation.
        :param offset: See gensim documentation.
        :param workers: Number of cores to use (if using multicore)
        :param use_multicore: Whether or not to use multicore

        **gensim_hdp**

        Fits an HDP model using the gensim implementation. Contrary to LDA and NMF, HDP attempts to auto-detect the
        correct number of topics. In practice, it actually fits ``T`` topics (default is 150) but many are extremely rare
        or occur only in a very few number of documents. To identify the topics that are actually useful, this function
        passes the original :py:class:`pandas.DataFrame` through the trained model after fitting, and identifies \
        topics that compose at least 1% of a document in at least 1% of all documents in the corpus. In other words, \
        topics are thrown out if the number of documents they appear in at a rate of at least 1% are fewer than 1% of \
        the total number of documents. Subsequent use of the model will only make use of topics that meet this \
        threshold. For more information on available parameters, please refer to the official documentation: \
        https://radimrehurek.com/gensim/models/hdpmodel.html

        :param df: The :py:class:`pandas.DataFrame` to train the model on (must contain ``self.text_col``)
        :param max_chunks: See gensim documentation.
        :param max_time: See gensim documentation.
        :param chunksize: See gensim documentation.
        :param kappa: See gensim documentation.
        :param tau: See gensim documentation.
        :param T: See gensim documentation.
        :param K: See gensim documentation.
        :param alpha: See gensim documentation.
        :param beta: See gensim documentation.
        :param gamma: See gensim documentation.
        :param scale: See gensim documentation.
        :param var_converge: See gensim documentation.

        **corex**

        Fits a CorEx topic model. Anchors can be provided in the form of a list of lists, with each item
        corresponding to a set of words to be used to seed a topic. For example:

        .. code-block:: python

            anchors=[
                ['cat', 'kitten'],
                ['dog', 'puppy']
            ]

        The list of anchors cannot be longer than the specified number of topics, and all of the words must
        exist in the vocabulary. The ``anchor_strength`` parameter determines the degree to which the model is able to
        override the suggested words based on the data; providing higher values are a way of "insisting" more strongly
        that the model keep the provided words together in a single topic. For more information on available \
        parameters, please refer to the official documentation: https://github.com/gregversteeg/corex_topic

        :param df: The :py:class:`pandas.DataFrame` to train the model on (must contain ``self.text_col``)
        :param anchors: A list of lists that contain words that the model should try to group together into topics
        :param anchor_strength: The degree to which the provided anchors should be preserved regardless of the data

        """

        fit_params = self.get_fit_params(**kwargs)

        if self.method in ["sklearn_lda", "sklearn_nmf"]:

            if self.method == "sklearn_lda":
                self.model = LatentDirichletAllocation(
                    n_components=self.num_topics, **fit_params
                )
            if self.method == "sklearn_nmf":
                self.model = NMF(n_components=self.num_topics, **fit_params)

            if is_not_null(df):
                features = self.get_features(df)
            else:
                features = self.train_features
            self.model.fit(features)

        elif self.method in ["gensim_lda", "gensim_hdp"]:

            vocab_dict = dict([(i, s) for i, s in enumerate(self.ngrams)])
            if is_not_null(df):
                features = self.get_features(df, keep_sparse=True)
            else:
                features = self.train_features
            matrix = gensim.matutils.Sparse2Corpus(features, documents_columns=False)

            if self.method == "gensim_lda":
                fit_params["num_topics"] = self.num_topics
                fit_params["id2word"] = vocab_dict
                if fit_params["use_multicore"]:
                    model_class = gensim.models.ldamulticore.LdaMulticore
                else:
                    model_class = gensim.models.LdaModel
                    del fit_params["workers"]
                del fit_params["use_multicore"]
                self.model = model_class(**fit_params)
                self.model.update(matrix)
            elif self.method == "gensim_hdp":
                model_class = gensim.models.hdpmodel.HdpModel
                self.model = model_class(matrix, vocab_dict, **fit_params)
                doc_topics = self.get_document_topics(self.df)
                topics = ((doc_topics >= 0.01).astype(int).mean() >= 0.01).astype(int)
                self.topic_ids = [
                    int(col.split("_")[-1])
                    for col in topics[topics == 1].index
                    if col.startswith("topic_")
                ]
                self.num_topics = len(self.topic_ids)

        elif self.method == "corex":

            if is_not_null(df):
                features = self.get_features(df, keep_sparse=True)
            else:
                features = self.get_features(self.train_df, keep_sparse=True)
            self.model = corextopic.Corex(n_hidden=self.num_topics)
            self.model.fit(features, words=self.ngrams, **fit_params)

    def get_score(self):

        """
        Returns goodness-of-fit scores for certain models, based on the holdout documents.

        .. note:: The following scores are available for the following methods:

                - perplexity: (sklearn_lda only) The model's perplexity
                - score: (sklearn_lda only) The model's log-likelihood score
                - total_correlation: (corex only) The model's total correlation score

        :return: A dictionary with goodness-of-fit scores
        :rtype: dict

        """

        if self.model:
            if self.method == "sklearn_lda":
                return {
                    "perplexity": self.model.perplexity(self.test_features),
                    "score": self.model.score(self.test_features),
                }
            elif self.method == "corex":
                return {"total_correlation": self.model.tc}
            else:
                return {}

    def get_document_topics(self, df, **kwargs):

        """
        Takes a :py:class:`pandas.DataFrame` and returns a document-topic :py:class:`pandas.DataFrame` \
        (rows=documents, columns=topics)

        :param df: The :py:class:`pandas.DataFrame` to process (must have ``self.text_col`` in it)
        :param min_probability: (gensim_lda use_multicore=False only) Topics with a probability lower than this \
        threshold will be filtered out (Default=0.0)
        :type min_probability: float
        :return: A document-topic matrix
        """

        if self.method in ["sklearn_lda", "sklearn_nmf"]:

            features = self.get_features(df)
            doc_topics = self.model.transform(features)
            topic_matrix = pd.DataFrame(
                doc_topics,
                columns=["topic_{}".format(i) for i in range(0, self.num_topics)],
                index=features.index,
            )
            return topic_matrix

        elif self.method in ["gensim_lda", "gensim_hdp"]:

            features = self.get_features(df, keep_sparse=True)
            matrix = gensim.matutils.Sparse2Corpus(features, documents_columns=False)
            rows = []
            for index, bow in zip(df.dropna(subset=[self.text_col]).index, matrix):
                if self.method == "gensim_lda":
                    if "min_probability" not in kwargs:
                        kwargs["min_probability"] = 0.0
                    try:
                        doc_topics = self.model.get_document_topics(bow, **kwargs)
                    except TypeError:
                        del kwargs["min_probability"]
                        doc_topics = self.model.get_document_topics(bow, **kwargs)
                elif self.method == "gensim_hdp":
                    doc_topics = self.model[bow]
                row = {"index": index}
                for topic, weight in doc_topics:
                    if self.method == "gensim_lda" or (
                        not self.topic_ids or topic in self.topic_ids
                    ):
                        row["topic_{}".format(topic)] = weight
                rows.append(row)
            df = pd.DataFrame(rows).fillna(0)
            df = df.set_index(df["index"])
            del df["index"]
            return df

        elif self.method == "corex":

            features = self.get_features(df, keep_sparse=True)
            doc_topics = self.model.transform(features)
            topic_matrix = pd.DataFrame(
                doc_topics,
                columns=["topic_{}".format(i) for i in range(0, self.num_topics)],
                index=df.index,
            )
            return topic_matrix

    def get_topics(self, include_weights=False, top_n=10, **kwargs):

        """
        Returns a list, equal in length to the number of topics, where each item is a list of words or word-weight
        tuples.

        :param include_weights: Whether or not to include weights along with the ngrams
        :type include_weights: bool
        :param top_n: The number of words to include for each topic
        :type top_n: init
        :return: A list of lists, where each item is a list of ngrams or ngram-weight tuples
        """

        if self.method in ["sklearn_lda", "sklearn_nmf"]:

            topic_features = self.model.components_
            topics = defaultdict(list)
            for topic_id, topic in enumerate(topic_features):
                top_ngram_index = sorted(
                    [
                        (ngram_id, float(ngram_value))
                        for ngram_id, ngram_value in enumerate(topic)
                    ],
                    key=lambda x: x[1],
                    reverse=True,
                )
                topics[topic_id] = [
                    self.ngrams[ngram_id]
                    if not include_weights
                    else (self.ngrams[ngram_id], ngram_value)
                    for ngram_id, ngram_value in top_ngram_index[:top_n]
                ]
            return topics

        elif self.method in ["gensim_lda", "gensim_hdp"]:

            topics = defaultdict(list)
            if self.method == "gensim_hdp":
                topic_ids = self.topic_ids
            else:
                topic_ids = range(self.num_topics)
            for i in topic_ids:
                for ngram, weight in self.model.show_topic(i, topn=top_n):
                    if include_weights:
                        topics[i].append((ngram, weight))
                    else:
                        topics[i].append(ngram)
            return topics

        elif self.method == "corex":

            topics = defaultdict(list)
            for topic_id, topic_ngrams in enumerate(
                self.model.get_topics(n_words=top_n)
            ):
                for ngram, weight, _ in topic_ngrams:
                    if include_weights:
                        topics[topic_id].append((ngram, weight))
                    else:
                        topics[topic_id].append(ngram)
            return topics

    def print_topics(self, include_weights=False, top_n=10):

        """
        Prints the top words for each topic from a trained model.

        :param include_weights: Whether or not to include weights along with the ngrams
        :type include_weights: bool
        :param top_n: The number of words to include for each topic
        :type top_n: int
        """

        for i, topic in self.get_topics(
            include_weights=include_weights, top_n=top_n
        ).items():
            print("{}: {}".format(i, ", ".join(topic)))



from __future__ import print_function

from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score


def compute_kmeans_clusters(features, k=10, return_score=False):
    """
    Uses K-Means to cluster an arbitrary set of features. This function expects input data where the rows are units \
    and columns are features.

    :param features: TF-IDF sparse matrix or :py:class:`pandas.DataFrame`
    :param k: The number of clusters to extract
    :type k: int
    :param return_score: If True, the function returns a tuple with the cluster \
    assignments and the silhouette score of the clustering; otherwise the function just returns a list of cluster \
    labels for each row. (Default=False)
    :type return_score: bool
    :return: A list with the cluster label for each row, or a tuple containing the \
    labels followed by the silhouette score of the K-Means model.
    :rtype: list

    Usage::

        from pewanalytics.stats.clustering import compute_kmeans_clusters
        from sklearn import datasets
        import pandas as pd

        # The iris dataset is a common example dataset included in scikit-learn with 3 main clusters
        # Let's see if we can find them
        df = pd.DataFrame(datasets.load_iris().data)

        >>> df['cluster'] = compute_kmeans_clusters(df, k=3)
        KMeans: n_clusters 3, score is 0.5576853964035263

        >>> df['cluster'].value_counts()
        1    62
        0    50
        2    38
        Name: cluster, dtype: int64

    """

    km = KMeans(n_clusters=k)
    labels = km.fit_predict(features)
    silhouette_avg = silhouette_score(features, labels)
    print("KMeans: n_clusters {}, score is {}".format(k, silhouette_avg))
    if not return_score:
        return km.labels_.tolist()
    else:
        return (km.labels_.tolist(), silhouette_avg)


def compute_hdbscan_clusters(features, min_cluster_size=100, min_samples=1, **kwargs):
    """
    Uses HDBSCAN* to identify the best number of clusters and map each unit to one. This function expects input data \
    where the rows are units and columns are features. Additional keyword arguments are passed to HDBSCAN. Check \
    out the official documentation for more: https://hdbscan.readthedocs.io/en/latest

    :param features: TF-IDF sparse matrix or :py:class:`pandas.DataFrame`
    :param min_cluster_size: int - minimum number of documents/units that can exist in a cluster.
    :type min_cluster_size: int
    :param min_samples: Minimum number of samples to draw (see HDBSCAN documentation for more)
    :type min_samples: int
    :param kwargs: Additional HDBSCAN parameters: https://hdbscan.readthedocs.io/en/latest/parameter_selection.html
    :return: A list with the cluster label for each row

    Usage::

        from pewanalytics.stats.clustering import compute_hdbscan_clusters
        from sklearn import datasets
        import pandas as pd

        df = pd.DataFrame(datasets.load_iris().data)

        >>> df['cluster'] = compute_hdbscan_clusters(df, min_cluster_size=10)
        HDBSCAN: n_clusters 2

        >>> df['cluster'].value_counts()
        1    100
        0     50
        Name: cluster, dtype: int64

    """

    import hdbscan

    clusterer = hdbscan.HDBSCAN(
        min_cluster_size=min_cluster_size, min_samples=min_samples, **kwargs
    )
    clusterer.fit(features)
    print("HDBSCAN: n_clusters {}".format(clusterer.labels_.max() + 1))
    return clusterer.labels_



from __future__ import print_function
from builtins import zip
from builtins import range
import pandas as pd

from mca import MCA
from sklearn.decomposition import PCA, TruncatedSVD


def _decompose(
    features, decompose_class, feature_names=None, k=20, component_prefix="component"
):
    """
    Internal function used to break apart a set of features using a scikit-learn decomposition class and return \
    the resulting matrices.

    :param features: A :py:class:`pandas.DataFrame` or sparse matrix with rows are documents and columns are features
    :param feature_names: An optional list of feature names (for sparse matrices)
    :type feature_names: list
    :param k: Number of dimensions to extract
    :type k: int
    :param component_prefix: A prefix for the column names
    :type component_prefix: str
    :return: A tuple of two :py:class:`pandas.DataFrame`s, (features x components, documents x components)
    :rtype: tuple
    """

    model = decompose_class(n_components=k)
    try:
        features = pd.DataFrame(features.todense())
    except:
        pass
    model.fit(features)
    score = sum(model.explained_variance_ratio_)
    print("Decomposition explained variance ratio: {}".format(score))
    if not feature_names:
        feature_names = features.columns
    components = pd.DataFrame(model.components_, columns=feature_names).transpose()
    print("Top features:")
    for col in components.columns:
        print(
            "Component {}: {}".format(
                col, components.sort_values(col, ascending=False)[:10].index.values
            )
        )
    components.columns = [
        "{}_{}".format(component_prefix, c) for c in components.columns
    ]
    results = pd.DataFrame(model.transform(features), index=features.index)
    results.columns = components.columns
    results[component_prefix] = results.idxmax(axis=1)
    return (components, results)


def get_pca(features, feature_names=None, k=20):

    """
    Performs PCA on a set of features. This function expects input data where the rows are units \
    and columns are features.

    For more information about how PCA is implemented, visit the \
    `Scikit-Learn Documentation <https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html>`__.

    :param features: A :py:class:`pandas.DataFrame` or sparse matrix where rows are units/observations and columns \
    are features
    :param feature_names: An optional list of feature names (for sparse matrices)
    :type feature_names: list
    :param k: Number of dimensions to extract
    :type k: int
    :return: A tuple of two :py:class:`pandas.DataFrame` s, (features x components, units x components)
    :rtype: tuple

    Usage::

        from pewanalytics.stats.dimensionality_reduction import get_pca
        from sklearn import datasets
        import pandas as pd

        df = pd.DataFrame(datasets.load_iris().data)

        >>> feature_weights, df_reduced  = get_pca(df, k=2)
        Decomposition explained variance ratio: 0.977685206318795
        Top features:
        Component 0: [2 0 3 1]
        Component 1: [1 0 3 2]

        >>> feature_weights
              pca_0     pca_1
        0  0.361387  0.656589
        1 -0.084523  0.730161
        2  0.856671 -0.173373
        3  0.358289 -0.075481

        >>> df_reduced.head()
              pca_0     pca_1    pca
        0 -2.684126  0.319397  pca_1
        1 -2.714142 -0.177001  pca_1
        2 -2.888991 -0.144949  pca_1
        3 -2.745343 -0.318299  pca_1
        4 -2.728717  0.326755  pca_1

    """

    return _decompose(
        features, PCA, feature_names=feature_names, k=k, component_prefix="pca"
    )


def get_lsa(features, feature_names=None, k=20):

    """
    Performs LSA on a set of features. This function expects input data where the rows are units \
    and columns are features.

    For more information about how LSA is implemented, visit the \
    `Scikit-Learn Documentation \
    <https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html>`__.

    :param features: A :py:class:`pandas.DataFrame` or sparse matrix with rows are units/observations and columns \
    are features
    :param feature_names: An optional list of feature names (for sparse matrices)
    :type feature_names: list
    :param k: Number of dimensions to extract
    :type k: int
    :return: A tuple of two :py:class:`pandas.DataFrame` s, (features x components, documents x components)
    :rtype: tuple

    Usage::

        from pewanalytics.stats.dimensionality_reduction import get_lsa
        from sklearn import datasets
        import pandas as pd

        df = pd.DataFrame(datasets.load_iris().data)

        >>> feature_weights, df_reduced  = get_lsa(df, k=2)
        Decomposition explained variance ratio: 0.9772093692426493
        Top features:
        Component 0: [0 2 1 3]
        Component 1: [1 0 3 2]

        >>> feature_weights
              lsa_0     lsa_1
        0  0.751108  0.284175
        1  0.380086  0.546745
        2  0.513009 -0.708665
        3  0.167908 -0.343671

        >>> df_reduced.head()
              lsa_0     lsa_1    lsa
        0  5.912747  2.302033  lsa_0
        1  5.572482  1.971826  lsa_0
        2  5.446977  2.095206  lsa_0
        3  5.436459  1.870382  lsa_0
        4  5.875645  2.328290  lsa_0

    """

    return _decompose(
        features, TruncatedSVD, feature_names=feature_names, k=k, component_prefix="lsa"
    )


def correspondence_analysis(edges, n=1):

    """
    Performs correspondence analysis on a set of features.

    Most useful in the context of network analysis, where you might wish to, for example, \
    identify the underlying dimension in a network of Twitter users by using a matrix representing whether \
    or not they follow one another (when news and political accounts are included, the \
    underlying dimension often appears to approximate the left-right political spectrum.)

    :param edges: A :py:class:`pandas.DataFrame` of NxN where both the rows and columns are "nodes" and the values \
    are some sort of closeness or similarity measure (like a cosine similarity matrix)
    :param n: The number of dimensions to extract
    :type n: int
    :return: A :py:class:`pandas.DataFrame` where rows are the units and the columns correspond to the extracted \
    dimensions.

    Usage::

        from pewanalytics.stats.dimensionality_reduction import correspondence_analysis
        import nltk
        import pandas as pd
        from sklearn.metrics.pairwise import linear_kernel
        from sklearn.feature_extraction.text import TfidfVectorizer

        nltk.download("inaugural")
        df = pd.DataFrame([
            {"speech": fileid, "text": nltk.corpus.inaugural.raw(fileid)} for fileid in nltk.corpus.inaugural.fileids()
        ])

        vec = TfidfVectorizer(min_df=10, max_df=.9).fit(df['text'])
        tfidf = vec.transform(df['text'])

        cosine_similarities = linear_kernel(tfidf)
        matrix = pd.DataFrame(cosine_similarities, columns=df['speech'])

        # Looks like the main source of variation in the language of inaugural speeches is time!

        >>> mca = correspondence_analysis(matrix)

        >>> mca.sort_values("mca_1").head()
                        node     mca_1
        57  1993-Clinton.txt -0.075508
        56    2017-Trump.txt -0.068168
        55  1997-Clinton.txt -0.061567
        54    1973-Nixon.txt -0.060698
        53     1989-Bush.txt -0.056305

        >>> mca.sort_values("mca_1").tail()
                       node     mca_1
        4    1877-Hayes.txt  0.040037
        3   1817-Monroe.txt  0.040540
        2     1845-Polk.txt  0.042847
        1   1849-Taylor.txt  0.050937
        0  1829-Jackson.txt  0.056201


    """

    mca_counts = MCA(edges)
    rows = []
    for r in sorted(
        zip(edges.columns, [m for m in mca_counts.fs_r(N=n)]),
        key=lambda x: x[1][0],
        reverse=True,
    ):
        row = {"node": r[0]}
        for i in range(n):
            try:
                row["mca_{}".format(i + 1)] = r[1][i]
            except:
                pass
        rows.append(row)
    mca = pd.DataFrame(rows)

    return mca



from builtins import object
import re
import datetime

from dateutil.parser import parse
from calendar import IllegalMonthError


class DateFinder(object):

    """
    A helper class to search for dates in text using a series of regular expressions and a parser from \
    :py:mod:`dateutil`. Verifies that :py:mod:`dateutil` did not auto-fill missing values in the date. Time \
    information will be automatically cleared out, but you can also pass a list of additional regular expression \
    patterns (as strings) to define other patterns that should be cleared out before scanning for dates.

    :param preprocessing_patterns: Optional list of additional patterns to clear out prior to searching for dates.
    :type preprocessing_patterns: list

    Usage::

        from pewanalytics.text.dates import DateFinder

        text = "January 1, 2018 and 02/01/2019 and Mar. 1st 2020"
        low_bound = datetime.datetime(2017, 1, 1)
        high_bound = datetime.datetime(2021, 1, 1)

        >>> finder = DateFinder()
        >>> dates = finder.find_dates(text, low_bound, high_bound)
        >>> dates
        [
            (datetime.datetime(2018, 1, 1, 0, 0), 'January 1, 2018 '),
            (datetime.datetime(2020, 3, 1, 0, 0), 'Mar. 1st 2020'),
            (datetime.datetime(2019, 2, 1, 0, 0), '02/01/2019 ')
        ]

    """

    def __init__(self, preprocessing_patterns=None):

        # A generally permissive date regex, also fairly prone to false positives
        self.date_regex = re.compile(
            r"""(?=((?ix)                       # case-insensitive, verbose regex
            \b                                  # match a word boundary
            (?:                                 # match the following three times:
             (?:                                # either
              \d+                               # a number,
              (?:\.|st|nd|rd|th|,)*             # followed by a dot, st, nd, rd, a comma, or th (optional)
              |                                 # or a month name
              (?:(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[a-z]*)
             )
             [\s./-]*                           # followed by a date separator or whitespace (optional)
            ){3}                                # do this three times
            \b)) """
        )

        # Always remove times, because those trip up dates
        self.preprocessing_patterns = [re.compile("((?:\d\d|\d):[0-9][0-9])")]
        # Add in any additional patterns to clear out, as provided by the user
        if preprocessing_patterns:
            self.preprocessing_patterns.extend(
                [re.compile(p) for p in preprocessing_patterns]
            )

    def _preprocess(self, text):

        """
        Return the text without any references to "parts" or specific times.

        :param text: A string to be cleaned that contains a date
        :type text: str
        :return: A cleaned string without additional time info and other boilerplate
        :rtype: str
        """

        for pattern in self.preprocessing_patterns:
            matches = re.findall(pattern, text)
            for match in matches:
                text = text.replace(match, "")

        return text

    def find_dates(self, text, cutoff_date_start, cutoff_date_end):

        """
        Return all of the dates (in text form and as datetime) in the text variable that fall within the specified \
        window of dates (inclusive).

        :param text: The text to scan for dates
        :type text: str
        :param cutoff_date_start: No dates will be returned if they fall before this date
        :type cutoff_date_start: `datetime.date`
        :param cutoff_date_end: No dates will be returned if they fall after this date
        :type cutoff_date_end: `datetime.date`
        :return: A list of tuples containing (datetime object, raw date text)
        :rtype: list
        """

        final_dates, suspected_dates = [], []

        # Start by stripping out time info, which can confuse the date regex and will lead to text snippets that \
        # can't be parsed into dates.
        text = self._preprocess(text)

        # Now find all of the plausible dates in the text. Many of these will be false positives.
        date_text_list = re.findall(self.date_regex, text)
        for date_raw_text in date_text_list:
            suspected_dates.append(date_raw_text)

        for date_raw_text in suspected_dates:

            # The dateutil parser fills in missing date components with the current day's date, or with a default \
            # date you can pass it. This is annoying, because partial dates could show up in bulk, leading to odd \
            # clumps of dates on the same date that you parsed. Our solution is to use the two cutoff dates as default \
            # dates, parsing each date twice. If the two dates are the same, then the default date was not used. \
            # The contents of these dates don't matter much. However, they cannot have a month, day OR year in common, \
            # otherwise this won't work.
            default_date_1 = datetime.datetime(day=1, year=2020, month=9)
            default_date_2 = datetime.datetime(day=2, year=1999, month=6)

            try:
                datetime_1 = parse(date_raw_text, fuzzy=True, default=default_date_1)
                datetime_2 = parse(date_raw_text, fuzzy=True, default=default_date_2)
                if (
                    datetime_1 == datetime_2
                    and cutoff_date_start <= datetime_1 <= cutoff_date_end
                ):
                    date = (datetime_1, date_raw_text)
                    final_dates.append(date)
            except (ValueError, IllegalMonthError):
                pass

        return list(set(final_dates))



import math
import copy
import scipy
import numpy as np
import pandas as pd
from nltk.metrics import masi_distance, jaccard_distance
from nltk.metrics.agreement import AnnotationTask
from sklearn.metrics import (
    matthews_corrcoef,
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    roc_auc_score,
    cohen_kappa_score,
)
from sklearn.preprocessing import MultiLabelBinarizer
from statsmodels.stats.weightstats import DescrStatsW
from statsmodels.stats.inter_rater import fleiss_kappa
from pewtils import is_not_null
from scipy import spatial


def kappa_sample_size_power(
    rate1, rate2, k1, k0, alpha=0.05, power=0.8, twosided=False
):

    """
    Python translation of the ``N.cohen.kappa`` function from the ``irr`` R package.

    Source: https://cran.r-project.org/web/packages/irr/irr.pdf

    :param rate1: The probability that the first rater will record a positive diagnosis
    :type rate1: float
    :param rate2: The probability that the second rater will record a positive diagnosis
    :type rate2: float
    :param k1: The true Cohen's Kappa statistic
    :type k1: float
    :param k0: The value of kappa under the null hypothesis
    :type k0: float
    :param alpha: Type I error of test
    :type alpha: float
    :param power: The desired power to detect the difference between true kappa and hypothetical kappa
    :type power: float
    :param twosided: Set this to ``True`` if the test is two-sided
    :param twosided: bool
    :return: Returns the required sample size
    :rtype: int

    """

    if not twosided:
        d = 1
    else:
        d = 2
    pi_rater1 = 1 - rate1
    pi_rater2 = 1 - rate2
    pie = rate1 * rate2 + pi_rater1 * pi_rater2
    pi0 = k1 * (1 - pie) + pie
    pi22 = (pi0 - rate1 + pi_rater2) / 2
    pi11 = pi0 - pi22
    pi12 = rate1 - pi11
    pi21 = rate2 - pi11
    pi0h = k0 * (1 - pie) + pie
    pi22h = (pi0h - rate1 + pi_rater2) / 2
    pi11h = pi0h - pi22h
    pi12h = rate1 - pi11h
    pi21h = rate2 - pi11h
    Q = (1 - pie) ** (-4) * (
        pi11 * (1 - pie - (rate2 + rate1) * (1 - pi0)) ** 2
        + pi22 * (1 - pie - (pi_rater2 + pi_rater1) * (1 - pi0)) ** 2
        + (1 - pi0) ** 2
        * (pi12 * (rate2 + pi_rater1) ** 2 + pi21 * (pi_rater2 + rate1) ** 2)
        - (pi0 * pie - 2 * pie + pi0) ** 2
    )
    Qh = (1 - pie) ** (-4) * (
        pi11h * (1 - pie - (rate2 + rate1) * (1 - pi0h)) ** 2
        + pi22h * (1 - pie - (pi_rater2 + pi_rater1) * (1 - pi0h)) ** 2
        + (1 - pi0h) ** 2
        * (pi12h * (rate2 + pi_rater1) ** 2 + pi21h * (pi_rater2 + rate1) ** 2)
        - (pi0h * pie - 2 * pie + pi0h) ** 2
    )
    N = (
        (
            scipy.stats.norm.ppf(1 - alpha / d) * math.sqrt(Qh)
            + scipy.stats.norm.ppf(power) * math.sqrt(Q)
        )
        / (k1 - k0)
    ) ** 2
    return math.ceil(N)


def kappa_sample_size_CI(kappa0, kappaL, props, kappaU=None, alpha=0.05):

    """

    Helps determine the required document sample size to confirm that Cohen's Kappa between coders is at or \
    above a minimum threhsold. Useful in situations where multiple coders code a set of documents for a binary outcome.

    This function takes the observed kappa and proportion of positive cases from the sample, along with a lower-bound \
    for the minimum acceptable kappa, and returns the sample size required to confirm that the coders' agreement is \
    truly above that minimum level of kappa with 95% certainty. If the current sample size is below the required \
    sample size returned by this function, it can provide a rough estimate of how many additional documents need \
    to be coded - assuming that the coders continue agreeing and observing positive cases at the same rate.

    Translated from the ``kappaSize`` R package, ``CIBinary``: \
    https://github.com/cran/kappaSize/blob/master/R/CIBinary.R

    :param kappa0: The preliminary value of kappa
    :param kappa0: float
    :param kappaL: The desired expected lower bound for a two-sided 100(1 - alpha) % confidence interval for kappa. \
    Alternatively, if kappaU is set to NA, the procedure produces the number of required subjects for a one-sided \
    confidence interval
    :type kappaL: float
    :param props: The anticipated prevalence of the desired trait
    :type props: float
    :param kappaU: The desired expected upper confidence limit for kappa
    :type kappaU: float
    :param alpha: The desired type I error rate
    :type alpha: float
    :return: Returns the required sample size

    Usage::

        from pewanalytics.stats.irr import kappa_sample_size_CI

        observed_kappa = 0.8
        desired_kappa = 0.7
        observed_proportion = 0.5

        >>> kappa_sample_size(observed_kappa, desired_kappa, observed_proportion)
        140

    """

    if not kappaU:
        chiCrit = scipy.stats.chi2.ppf((1 - 2 * alpha), 1)

    if kappaL and kappaU:
        chiCrit = scipy.stats.chi2.ppf((1 - alpha), 1)

    def CalcIT(rho0, rho1, Pi, n):
        def P0(r, p):
            x = (1 - p) ** 2 + r * p * (1 - p)
            return x

        def P1(r, p):
            x = 2 * (1 - r) * p * (1 - p)
            return x

        def P2(r, p):
            x = p ** 2 + r * p * (1 - p)
            return x

        r1 = ((n * P0(r=rho0, p=Pi)) - (n * P0(r=rho1, p=Pi))) ** 2 / (
            n * P0(r=rho1, p=Pi)
        )
        r2 = ((n * P1(r=rho0, p=Pi)) - (n * P1(r=rho1, p=Pi))) ** 2 / (
            n * P1(r=rho1, p=Pi)
        )
        r3 = ((n * P2(r=rho0, p=Pi)) - (n * P2(r=rho1, p=Pi))) ** 2 / (
            n * P2(r=rho1, p=Pi)
        )
        return sum([r for r in [r1, r2, r3] if r])

    n = 10
    result = 0
    while (result - 0.001) < chiCrit:
        result = CalcIT(kappa0, kappaL, props, n)
        n += 1
    return n


def compute_scores(
    coder_df,
    coder1,
    coder2,
    outcome_column,
    document_column,
    coder_column,
    weight_column=None,
    pos_label=None,
):

    """
    Computes a variety of inter-rater reliability scores, including Cohen's kappa, Krippendorf's alpha, precision,
    and recall. The input data must consist of a :py:class:`pandas.DataFrame` with the following columns:

        - A column with values that indicate the coder (like a name)
        - A column with values that indicate the document (like an ID)
        - A column with values that indicate the code value
        - (Optional) A column with document weights

    This function will return a :py:class:`pandas.DataFrame` with agreement scores between the two specified coders.

    :param coder_df: A :py:class:`pandas.DataFrame` of codes
    :type coder_df: :py:class:`pandas.DataFrame`
    :param coder1: The value in ``coder_column`` for rows corresponding to the first coder
    :type coder1: str or int
    :param coder2: The value in ``coder_column`` for rows corresponding to the second coder
    :type coder2: str or int
    :param outcome_column: The column that contains the codes
    :type outcome_column: str
    :param document_column: The column that contains IDs for the documents
    :type document_column: str
    :param coder_column: The column containing values that indicate which coder assigned the code
    :type coder_column: str
    :param weight_column: The column that contains sampling weights
    :type weight_column: str
    :param pos_label: The value indicating a positive label (optional)
    :type pos_label: str or int
    :return: A dictionary of scores
    :rtype: dict

    .. note:: If using a multi-class (non-binary) code, some scores may come back null or not compute as expected. \
        We recommend running the function separately for each specific code value as a binary flag by providing \
        each unique value to the ``pos_label`` argument. If ``pos_label`` is not provided for multi-class codes, \
        this function will attempt to compute scores based on support-weighted averages.

    Usage::

        from pewanalytics.stats.irr import compute_scores
        import pandas as pd

        df = pd.DataFrame([
            {"coder": "coder1", "document": 1, "code": "2"},
            {"coder": "coder2", "document": 1, "code": "2"},
            {"coder": "coder1", "document": 2, "code": "1"},
            {"coder": "coder2", "document": 2, "code": "2"},
            {"coder": "coder1", "document": 3, "code": "0"},
            {"coder": "coder2", "document": 3, "code": "0"},
        ])

        >>> compute_scores(df, "coder1", "coder2", "code", "document", "coder")
        {'coder1': 'coder1',
         'coder2': 'coder2',
         'n': 3,
         'outcome_column': 'code',
         'pos_label': None,
         'coder1_mean_unweighted': 1.0,
         'coder1_std_unweighted': 0.5773502691896257,
         'coder2_mean_unweighted': 1.3333333333333333,
         'coder2_std_unweighted': 0.6666666666666666,
         'alpha_unweighted': 0.5454545454545454,
         'accuracy': 0.6666666666666666,
         'f1': 0.5555555555555555,
         'precision': 0.5,
         'recall': 0.6666666666666666,
         'precision_recall_min': 0.5,
         'matthews_corrcoef': 0.6123724356957946,
         'roc_auc': None,
         'pct_agree_unweighted': 0.6666666666666666}

        >>> compute_scores(df, "coder1", "coder2", "code", "document", "coder", pos_label="0")
         {'coder1': 'coder1',
         'coder2': 'coder2',
         'n': 3,
         'outcome_column': 'code',
         'pos_label': '0',
         'coder1_mean_unweighted': 0.3333333333333333,
         'coder1_std_unweighted': 0.3333333333333333,
         'coder2_mean_unweighted': 0.3333333333333333,
         'coder2_std_unweighted': 0.3333333333333333,
         'alpha_unweighted': 1.0,
         'cohens_kappa': 1.0,
         'accuracy': 1.0,
         'f1': 1.0,
         'precision': 1.0,
         'recall': 1.0,
         'precision_recall_min': 1.0,
         'matthews_corrcoef': 1.0,
         'roc_auc': 1.0,
         'pct_agree_unweighted': 1.0}

        >>> compute_scores(df, "coder1", "coder2", "code", "document", "coder", pos_label="1")
        {'coder1': 'coder1',
         'coder2': 'coder2',
         'n': 3,
         'outcome_column': 'code',
         'pos_label': '1',
         'coder1_mean_unweighted': 0.3333333333333333,
         'coder1_std_unweighted': 0.3333333333333333,
         'coder2_mean_unweighted': 0.0,
         'coder2_std_unweighted': 0.0,
         'alpha_unweighted': 0.0,
         'cohens_kappa': 0.0,
         'accuracy': 0.6666666666666666,
         'f1': 0.0,
         'precision': 0.0,
         'recall': 0.0,
         'precision_recall_min': 0.0,
         'matthews_corrcoef': 1.0,
         'roc_auc': None,
         'pct_agree_unweighted': 0.6666666666666666}

        >>> compute_scores(df, "coder1", "coder2", "code", "document", "coder", pos_label="2")
        {'coder1': 'coder1',
         'coder2': 'coder2',
         'n': 3,
         'outcome_column': 'code',
         'pos_label': '2',
         'coder1_mean_unweighted': 0.3333333333333333,
         'coder1_std_unweighted': 0.3333333333333333,
         'coder2_mean_unweighted': 0.6666666666666666,
         'coder2_std_unweighted': 0.3333333333333333,
         'alpha_unweighted': 0.4444444444444444,
         'cohens_kappa': 0.3999999999999999,
         'accuracy': 0.6666666666666666,
         'f1': 0.6666666666666666,
         'precision': 0.5,
         'recall': 1.0,
         'precision_recall_min': 0.5,
         'matthews_corrcoef': 0.5,
         'roc_auc': 0.75,
         'pct_agree_unweighted': 0.6666666666666666}


    """

    old_np_settings = np.seterr(all="raise")

    coder_df = copy.deepcopy(coder_df)
    if pos_label:
        coder_df[outcome_column] = (coder_df[outcome_column] == pos_label).astype(int)
    coder1_df = coder_df[coder_df[coder_column] == coder1]
    coder1_df.index = coder1_df[document_column]
    coder2_df = coder_df[coder_df[coder_column] == coder2]
    coder2_df.index = coder2_df[document_column]
    coder1_df = coder1_df[coder1_df.index.isin(coder2_df.index)]
    coder2_df = coder2_df[coder2_df.index.isin(coder1_df.index)].loc[coder1_df.index]

    row = {
        "coder1": coder1,
        "coder2": coder2,
        "n": len(coder1_df),
        "outcome_column": outcome_column,
        "pos_label": pos_label,
    }

    for labelsetname, labelset in [
        ("coder1", coder1_df[outcome_column]),
        ("coder2", coder2_df[outcome_column]),
    ]:

        if weight_column:
            try:
                weighted_stats = DescrStatsW(labelset, weights=coder1_df[weight_column])
                if weighted_stats:
                    row["{}_mean".format(labelsetname)] = weighted_stats.mean
                    row["{}_std".format(labelsetname)] = weighted_stats.std_mean
            except (TypeError, ValueError):
                try:
                    weighted_stats = DescrStatsW(
                        labelset.astype(int), weights=coder1_df[weight_column]
                    )
                    if weighted_stats:
                        row["{}_mean".format(labelsetname)] = weighted_stats.mean
                        row["{}_std".format(labelsetname)] = weighted_stats.std_mean
                except (TypeError, ValueError):
                    pass

        try:
            unweighted_stats = DescrStatsW(labelset, weights=[1.0 for x in labelset])
            if unweighted_stats:
                row["{}_mean_unweighted".format(labelsetname)] = unweighted_stats.mean
                row[
                    "{}_std_unweighted".format(labelsetname)
                ] = unweighted_stats.std_mean
        except (TypeError, ValueError):
            try:
                unweighted_stats = DescrStatsW(
                    labelset.astype(int), weights=[1.0 for x in labelset]
                )
                if unweighted_stats:
                    row[
                        "{}_mean_unweighted".format(labelsetname)
                    ] = unweighted_stats.mean
                    row[
                        "{}_std_unweighted".format(labelsetname)
                    ] = unweighted_stats.std_mean
            except (TypeError, ValueError):
                pass

    alpha = AnnotationTask(
        data=coder_df[[coder_column, document_column, outcome_column]].values
    )
    try:
        alpha = alpha.alpha()
    except (ZeroDivisionError, ValueError):
        alpha = None
    row["alpha_unweighted"] = alpha

    labels = np.unique(coder_df[outcome_column])
    if len(labels) <= 2:

        try:
            row["cohens_kappa"] = cohen_kappa_score(
                coder1_df[outcome_column],
                coder2_df[outcome_column],
                sample_weight=coder1_df[weight_column] if weight_column else None,
                labels=labels,
            )
        except FloatingPointError:
            row["cohens_kappa"] = 1.0

    try:
        row["accuracy"] = accuracy_score(
            coder1_df[outcome_column],
            coder2_df[outcome_column],
            sample_weight=coder1_df[weight_column] if weight_column else None,
        )
    except ValueError:
        row["accuracy"] = None

    try:
        row["f1"] = f1_score(
            coder1_df[outcome_column],
            coder2_df[outcome_column],
            sample_weight=coder1_df[weight_column] if weight_column else None,
            labels=labels,
            average="weighted" if not pos_label else "binary",
        )
    except ValueError:
        row["f1"] = None

    try:
        row["precision"] = precision_score(
            coder1_df[outcome_column],
            coder2_df[outcome_column],
            sample_weight=coder1_df[weight_column] if weight_column else None,
            labels=labels,
            average="weighted" if not pos_label else "binary",
        )
    except ValueError:
        row["precision"] = None

    try:
        row["recall"] = recall_score(
            coder1_df[outcome_column],
            coder2_df[outcome_column],
            sample_weight=coder1_df[weight_column] if weight_column else None,
            labels=labels,
            average="weighted" if not pos_label else "binary",
        )
    except ValueError:
        row["recall"] = None

    if is_not_null(row["precision"]) and is_not_null(row["recall"]):
        row["precision_recall_min"] = min([row["precision"], row["recall"]])
    else:
        row["precision_recall_min"] = None

    try:
        row["matthews_corrcoef"] = matthews_corrcoef(
            coder1_df[outcome_column],
            coder2_df[outcome_column],
            sample_weight=coder1_df[weight_column] if weight_column else None,
        )
    except ValueError:
        row["matthews_corrcoef"] = None
    except FloatingPointError:
        row["matthews_corrcoef"] = 1.0
    if row["accuracy"] == 1.0:
        # In newer versions of sklearn, perfect agreement returns zero incorrectly
        # We'll use the accuracy score to detect perfect agreement and correct it:
        row["matthews_corrcoef"] = 1.0

    try:

        row["roc_auc"] = (
            roc_auc_score(
                coder1_df[outcome_column],
                coder2_df[outcome_column],
                sample_weight=coder1_df[weight_column] if weight_column else None,
                labels=labels,
                average="weighted" if not pos_label else None,
            )
            if len(np.unique(coder1_df[outcome_column])) > 1
            and len(np.unique(coder2_df[outcome_column])) > 1
            else None
        )
    except TypeError:
        try:
            row["roc_auc"] = (
                roc_auc_score(
                    coder1_df[outcome_column],
                    coder2_df[outcome_column],
                    sample_weight=coder1_df[weight_column] if weight_column else None,
                    average="weighted" if not pos_label else None,
                )
                if len(np.unique(coder1_df[outcome_column])) > 1
                and len(np.unique(coder2_df[outcome_column])) > 1
                else None
            )
        except (ValueError, TypeError):
            row["roc_auc"] = None
    except (ValueError, TypeError):
        row["roc_auc"] = None

    row["pct_agree_unweighted"] = np.average(
        [
            1 if c[0] == c[1] else 0
            for c in zip(coder1_df[outcome_column], coder2_df[outcome_column])
        ]
    )

    for k, v in row.items():
        if type(v) == tuple:
            row[k] = v[0]
            # For some weird reason, some of the sklearn scorers return 1-tuples sometimes

    np.seterr(**old_np_settings)

    return row


def compute_overall_scores(coder_df, outcome_column, document_column, coder_column):

    """
    Computes overall inter-rater reliability scores (Krippendorf's Alpha and Fleiss' Kappa). Allows for more than two \
    coders and code values. The input data must consist of a :py:class:`pandas.DataFrame` with the following columns:

        - A column with values that indicate the coder (like a name)
        - A column with values that indicate the document (like an ID)
        - A column with values that indicate the code value

    :param coder_df: A :py:class:`pandas.DataFrame` of codes
    :type coder_df: :py:class:`pandas.DataFrame`
    :param outcome_column: The column that contains the codes
    :type outcome_column: str
    :param document_column: The column that contains IDs for the documents
    :type document_column: str
    :param coder_column: The column containing values that indicate which coder assigned the code
    :type coder_column: str
    :return: A dictionary containing the scores
    :rtype: dict

    Usage::

        from pewanalytics.stats.irr import compute_overall_scores
        import pandas as pd

        df = pd.DataFrame([
            {"coder": "coder1", "document": 1, "code": "2"},
            {"coder": "coder2", "document": 1, "code": "2"},
            {"coder": "coder1", "document": 2, "code": "1"},
            {"coder": "coder2", "document": 2, "code": "2"},
            {"coder": "coder1", "document": 3, "code": "0"},
            {"coder": "coder2", "document": 3, "code": "0"},
        ])

        >>> compute_overall_scores(df, "code", "document", "coder")
        {'alpha': 0.5454545454545454, 'fleiss_kappa': 0.4545454545454544}

    """

    alpha = AnnotationTask(
        data=coder_df[[coder_column, document_column, outcome_column]].values
    )
    try:
        alpha = alpha.alpha()
    except (ZeroDivisionError, ValueError):
        alpha = None

    grouped = coder_df.groupby(document_column).count()
    complete_docs = grouped[
        grouped[coder_column] == len(coder_df[coder_column].unique())
    ].index
    dataset = coder_df[coder_df[document_column].isin(complete_docs)]
    df = dataset.groupby([outcome_column, document_column]).count()[[coder_column]]
    df = df.unstack(outcome_column).fillna(0)

    if len(df) > 0:
        kappa = fleiss_kappa(df)
    else:
        kappa = None

    return {"alpha": alpha, "fleiss_kappa": kappa}


def compute_overall_scores_multivariate(
    coder_df, document_column, coder_column, outcome_columns
):
    """
    Computes overall inter-rater reliability scores (Krippendorf's Alpha and Fleiss' Kappa). Allows for more than two \
    coders, code values, AND variables. All variables and values will be converted into a matrix of dummy variables, \
    and Alpha and Kappa will be computed using four different distance metrics:

        - Discrete agreement (exact agreement across all outcome columns)
        - Jaccard coefficient
        - MASI distance
        - Cosine similarity

    The input data must consist of a :py:class:`pandas.DataFrame` with the following \
    columns:

        - A column with values that indicate the coder (like a name)
        - A column with values that indicate the document (like an ID)
        - One or more columns with values that indicate code values

    This code was adapted from a very helpful StackExchange post:
    https://stats.stackexchange.com/questions/511927/interrater-reliability-with-multi-rater-multi-label-dataset

    :param coder_df: A :py:class:`pandas.DataFrame` of codes
    :type coder_df: :py:class:`pandas.DataFrame`
    :param document_column: The column that contains IDs for the documents
    :type document_column: str
    :param coder_column: The column containing values that indicate which coder assigned the code
    :type coder_column: str
    :param outcome_columns: The columns that contains the codes
    :type outcome_columns: list
    :return: A dictionary containing the scores
    :rtype: dict

    Usage::

        from pewanalytics.stats.irr import compute_overall_scores_multivariate
        import pandas as pd

        coder_df = pd.DataFrame([
            {"coder": "coder1", "document": 1, "code": "2"},
            {"coder": "coder2", "document": 1, "code": "2"},
            {"coder": "coder1", "document": 2, "code": "1"},
            {"coder": "coder2", "document": 2, "code": "2"},
            {"coder": "coder1", "document": 3, "code": "0"},
            {"coder": "coder2", "document": 3, "code": "0"},
        ])

        >>> compute_overall_scores_multivariate(coder_df, 'document', 'coder', ["code"])
        {'fleiss_kappa_discrete': 0.4545454545454544,
         'fleiss_kappa_jaccard': 0.49999999999999994,
         'fleiss_kappa_masi': 0.49999999999999994,
         'fleiss_kappa_cosine': 0.49999999999999994,
         'alpha_discrete': 0.5454545454545454,
         'alpha_jaccard': 0.5454545454545454,
         'alpha_masi': 0.5454545454545454,
         'alpha_cosine': 0.5454545454545454}

        coder_df = pd.DataFrame([
            {"coder": "coder1", "document": 1, "code1": "2", "code2": "1"},
            {"coder": "coder2", "document": 1, "code1": "2", "code2": "1"},
            {"coder": "coder1", "document": 2, "code1": "1", "code2": "0"},
            {"coder": "coder2", "document": 2, "code1": "2", "code2": "1"},
            {"coder": "coder1", "document": 3, "code1": "0", "code2": "0"},
            {"coder": "coder2", "document": 3, "code1": "0", "code2": "0"},
        ])

        >>> compute_overall_scores_multivariate(coder_df, 'document', 'coder', ["code1", "code2"])
        {'fleiss_kappa_discrete': 0.4545454545454544,
         'fleiss_kappa_jaccard': 0.49999999999999994,
         'fleiss_kappa_masi': 0.49999999999999994,
         'fleiss_kappa_cosine': 0.49999999999999994,
         'alpha_discrete': 0.5454545454545454,
         'alpha_jaccard': 0.5161290322580645,
         'alpha_masi': 0.5361781076066792,
         'alpha_cosine': 0.5}

    """

    dummies = pd.concat(
        [pd.get_dummies(coder_df[c], prefix=c) for c in outcome_columns], axis=1
    )
    outcome_columns = dummies.columns
    coder_df = pd.concat([coder_df, dummies], axis=1)
    coder_df = coder_df.dropna(subset=outcome_columns)
    coder_df["value"] = coder_df.apply(
        lambda x: frozenset([col for col in outcome_columns if int(x[col]) == 1]),
        axis=1,
    )

    discrete_task = AnnotationTask(
        data=coder_df[[coder_column, document_column, "value"]].values
    )
    kappa_discrete = discrete_task.multi_kappa()
    alpha_discrete = discrete_task.alpha()

    task_data = coder_df[[coder_column, document_column, "value"]].values
    task_data = [tuple(t) for t in task_data]

    jaccard_task = AnnotationTask(data=task_data, distance=jaccard_distance)
    masi_task = AnnotationTask(data=task_data, distance=masi_distance)

    mlb = MultiLabelBinarizer()
    task_data = [
        (r[0][0], r[0][1], tuple(r[1]))
        for r in zip(task_data, mlb.fit_transform([t[2] for t in task_data]))
    ]
    cosine_task = AnnotationTask(data=task_data, distance=spatial.distance.cosine)

    kappa_jaccard = jaccard_task.multi_kappa()
    kappa_masi = masi_task.multi_kappa()
    kappa_cosine = cosine_task.multi_kappa()
    alpha_jaccard = jaccard_task.alpha()
    alpha_masi = masi_task.alpha()
    alpha_cosine = cosine_task.alpha()

    return {
        "fleiss_kappa_discrete": kappa_discrete,
        "fleiss_kappa_jaccard": kappa_jaccard,
        "fleiss_kappa_masi": kappa_masi,
        "fleiss_kappa_cosine": kappa_cosine,
        "alpha_discrete": alpha_discrete,
        "alpha_jaccard": alpha_jaccard,
        "alpha_masi": alpha_masi,
        "alpha_cosine": alpha_cosine,
    }



from builtins import str
from builtins import object
import nltk
import spacy
import re
from collections import defaultdict
from nltk.corpus import stopwords
from nltk import word_tokenize, pos_tag, ne_chunk
from pewtils import decode_text


class NamedEntityExtractor(object):

    """

    A wrapper around NLTK and SpaCy for named entity extraction. May be expanded to include more libraries in the \
    future.

    :param method: Specify the library to use when extracting methods. Options are 'nltk', 'spacy', 'all'. If \
    'all' is selected, both libraries will be used and the union will be returned. (Default='spacy')
    :type method: str

    Usage::

        from pewanalytics.text.ner import NamedEntityExtractor
        import nltk

        nltk.download("inaugural")
        fileid = nltk.corpus.inaugural.fileids()[0]
        text = nltk.corpus.inaugural.raw(fileid)

        >>> ner = NamedEntityExtractor(method="nltk")
        >>> ner.extract(text)
        {
            'ORGANIZATION': [
                'Parent', 'Invisible Hand', 'Great Author', 'House', 'Constitution', 'Senate',
                'Human Race', 'Representatives'
            ],
            'PERSON': ['Almighty Being'],
            'GPE': ['Heaven', 'United States', 'American']
        }

        >>> ner = NamedEntityExtractor(method="spacy")
        >>> ner.extract(text)
        {
            'ORGANIZATION': ['House of Representatives', 'Senate', 'Parent of the Human Race'],
            'DATE': ['present month', 'every day', '14th day', 'years'],
            'ORDINAL': ['first', 'fifth'],
            'GPE': ['United States'],
            'NORP': ['republican', 'American'],
            'LAW': ['Constitution']
        }

        >>> ner = NamedEntityExtractor(method="all")
        >>> ner.extract(text)
        {
            'ORGANIZATION': [
                'Representatives', 'Great Author', 'House', 'Parent', 'House of Representatives',
                'Parent of the Human Race', 'Invisible Hand', 'Human Race', 'Senate', 'Constitution'
            ],
            'PERSON': ['Almighty Being'],
            'GPE': ['Heaven', 'United States', 'American'],
            'DATE': ['every day', 'present month', '14th day', 'years'],
            'ORDINAL': ['first', 'fifth'],
            'NORP': ['republican', 'American'],
            'LAW': ['Constitution']
        }

    """

    def __init__(self, method="spacy"):

        if method not in ["nltk", "spacy", "all"]:
            raise Exception("Available methods are: 'nltk', 'spacy', 'all'")
        self.method = method

        self.type_map = {
            "ORG": "ORGANIZATION",
            "PER": "PERSON",
            "LOC": "LOCATION",
            "FAC": "FACILITY",
            "VEH": "VEHICLE",
            "WEA": "WEAPON",
            "GSP": "GPE",
        }

    def extract(self, text):

        """
        :param text: a string from which to extract named entities
        :type text: str
        :return: dictionary of entities organized by their category
        :rtype: dict
        """

        try:
            text = str(text)
        except Exception as e:
            text = decode_text(text)

        roots = defaultdict(list)

        if self.method in ["nltk", "all"]:

            try:
                tree = ne_chunk(pos_tag(word_tokenize(text)))
            except LookupError:
                nltk.download("maxent_ne_chunker")
                nltk.download("words")
                tree = ne_chunk(pos_tag(word_tokenize(text)), binary=True)

            for branch in tree:
                if type(branch) is nltk.Tree:
                    leaf = [" ".join(x[0] for x in branch.leaves())]
                    key = self.type_map.get(branch.label(), branch.label())
                    roots[key].extend(leaf)

        if self.method in ["spacy", "all"]:

            # SpaCy
            try:
                nlp = spacy.load("en_core_web_sm")
            except OSError:
                spacy.cli.download("en_core_web_sm")
                nlp = spacy.load("en_core_web_sm")
            for entity in nlp(text).ents:
                entity_text = re.sub(
                    r"^({})\s".format("|".join(stopwords.words("english"))),
                    "",
                    entity.text,
                )
                key = self.type_map.get(entity.label_, entity.label_)
                roots[key].append(entity_text)

        return {self.type_map.get(k, k): list(set(v)) for k, v in roots.items()}



from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import range
from builtins import object
import random
import numpy as np


def compute_sample_weights_from_frame(frame, sample, weight_vars):

    """
    Takes two :py:class:`pandas.DataFrame` s and computes sampling weights for the second one, based on the first. \
    The first :py:class:`pandas.DataFrame` should be equivalent to the population that the second \
    :py:class:`pandas.DataFrame`, a sample, was drawn from. Weights will be calculated based on the differences in \
    the distribution of one or more variables specified in ``weight_vars`` (these should be the names of columns). \
    Returns a :py:class:`pandas.Series` equal in length to the ``sample`` with the computed weights.

    :param frame: :py:class:`pandas.DataFrame` (must contain all of the columns specified in ``weight_vars``)
    :param sample: :py:class:`pandas.DataFrame` (must contain all of the columns specified in ``weight_vars``)
    :param weight_vars: The names of the columns to use when computing weights.
    :type weight_vars: list
    :return: A :py:class:`pandas.Series` containing the weights for each row in the ``sample``

    Usage::

        from pewanalytics.stats.sampling import compute_sample_weights_from_frame
        import nltk
        import pandas as pd
        from sklearn.metrics.pairwise import linear_kernel
        from sklearn.feature_extraction.text import TfidfVectorizer

        nltk.download("inaugural")
        frame = pd.DataFrame([
            {"speech": fileid, "text": nltk.corpus.inaugural.raw(fileid)} for fileid in nltk.corpus.inaugural.fileids()
        ])
        # Let's grab a sample of speeches - some that mention specific terms, and an additional random sample
        frame['economy'] = frame['text'].str.contains("economy").astype(int)
        frame['health'] = frame['text'].str.contains("health").astype(int)
        frame['immigration'] = frame['text'].str.contains("immigration").astype(int)
        frame['education'] = frame['text'].str.contains("education").astype(int)
        sample = pd.concat([
            frame[frame['economy']==1].sample(5),
            frame[frame['health']==1].sample(5),
            frame[frame['immigration']==1].sample(5),
            frame[frame['education']==1].sample(5),
            frame.sample(5)
        ])
        # Now we can get the sampling weights to adjust it back to the population based on those variables

        >>> sample['weight'] = compute_sample_weights_from_frame(frame, sample, ["economy", "health", "immigration", "education"])
        >>> sample
                       speech                                               text  economy  health  immigration  education  count    weight
        7     1817-Monroe.txt  I should be destitute of feeling if I was not ...        1       1            0          0      1  1.005747
        11   1833-Jackson.txt  Fellow citizens, the will of the American peop...        1       0            0          0      1  2.370690
        34  1925-Coolidge.txt  My countrymen, no one can contemplate curre...           1       0            1          1      1  0.344828
        35    1929-Hoover.txt  My Countrymen: This occasion is not alone the ...        1       1            0          1      1  0.538793
        28  1901-McKinley.txt  My fellow-citizens, when we assembled here on ...        1       0            0          0      1  2.370690

    """

    if len(weight_vars) > 0:

        frame["count"] = 1
        sample["count"] = 1
        sample_grouped = sample.groupby(weight_vars).count()
        sample_grouped /= len(sample)
        frame_grouped = frame.groupby(weight_vars).count()
        frame_grouped /= len(frame)
        weights = frame_grouped / sample_grouped
        weights["weight"] = weights["count"]
        for c in weights.columns:
            if c not in weight_vars and c != "weight":
                del weights[c]
        try:
            sample = sample.merge(
                weights, how="left", left_on=weight_vars, right_index=True
            )
        except ValueError:
            weights = weights.reset_index()
            index = sample.index
            sample = sample.merge(
                weights, how="left", left_on=weight_vars, right_on=weight_vars
            )
            sample.index = index
    else:
        sample["weight"] = 1.0

    return sample["weight"]


def compute_balanced_sample_weights(sample, weight_vars, weight_column=None):

    """
    Takes a :py:class:`pandas.DataFrame` and one or more column names (``weight_vars``) and computes weights such \
    that every unique combination of values in the weighting columns are balanced (when weighted, the sum of the \
    observations with each combination will be equal to one another). Useful for balancing important groups in \
    training datasets, etc.

    :param sample: :py:class:`pandas.DataFrame` (must contain all of the columns specified in ``weight_vars``)
    :param weight_vars: The names of the columns to use when computing weights.
    :type weight_vars: list
    :param weight_column: An option column containing existing weights, which can be factored into the new weights.
    :type weight_column: str
    :return: A :py:class:`pandas.Series` containing the weights for each row in the ``sample``

    .. note:: All weight variables must be binary flags (1 or 0); if you want to weight using a non-binary variable, \
        you should convert it into a set of dummy variables and then pass those in as multiple columns.
        
    Usage::

        from pewanalytics.stats.sampling import compute_balanced_sample_weights
        import pandas as pd

        # Let's say we have a set of tweets from members of Congress
        df = pd.DataFrame([
            {"politician_id": 1, "party": "R", "tweet": "Example document"},
            {"politician_id": 1, "party": "R", "tweet": "Example document"},
            {"politician_id": 2, "party": "D", "tweet": "Example document"},
            {"politician_id": 2, "party": "D", "tweet": "Example document"},
            {"politician_id": 3, "party": "D", "tweet": "Example document"},
        ])
        df['is_republican'] = (df['party']=="R").astype(int)

        # We can balance the parties like so:

        >>> df['weight'] = compute_balanced_sample_weights(df, ["is_republican"])

        >>> df
           politician_id party             tweet  is_rep    weight  is_republican
        0              1     R  Example document       1  1.250000              1
        1              1     R  Example document       1  1.250000              1
        2              2     D  Example document       0  0.833333              0
        3              2     D  Example document       0  0.833333              0
        4              3     D  Example document       0  0.833333              0

    """

    if len(weight_vars) > 0:

        num_valid_combos = 0
        weight_vars = list(set(weight_vars))
        combo_weights = {}
        combos = list(
            set(
                [
                    tuple(row[weight_vars].values.astype(bool))
                    for index, row in sample.iterrows()
                ]
            )
        )
        for combo in combos:
            if weight_column:
                combo_weights[combo] = float(
                    sample[
                        eval(
                            " & ".join(
                                [
                                    "(sample['{}']=={})".format(col, c)
                                    for col, c in zip(weight_vars, combo)
                                ]
                            )
                        )
                    ][weight_column].sum()
                ) / float(sample[weight_column].sum())
            else:
                combo_weights[combo] = float(
                    len(
                        sample[
                            eval(
                                " & ".join(
                                    [
                                        "(sample['{}']=={})".format(col, c)
                                        for col, c in zip(weight_vars, combo)
                                    ]
                                )
                            )
                        ]
                    )
                ) / float(len(sample))
            if combo_weights[combo] > 0:
                num_valid_combos += 1
            else:
                del combo_weights[combo]

        balanced_ratio = 1.0 / float(num_valid_combos)
        combo_weights = {
            k: float(balanced_ratio) / float(v) for k, v in combo_weights.items()
        }

        sample["weight"] = sample.apply(
            lambda x: combo_weights[tuple([x[v] for v in weight_vars])], axis=1
        )

    else:
        sample["weight"] = 1.0

    return sample["weight"]


class SampleExtractor(object):
    """
    A helper class for extracting samples using various sampling methods.

    :param df: The sampling frame
    :type df: :py:class:`pandas.DataFrame`
    :param id_col: Column in the :py:class:`pandas.DataFrame` to be used as the unique ID of observations
    :type id_col: str
    :param verbose: Whether or not to print information during the sampling process (default=False)
    :type verbose: bool
    :param seed: Random seed (optional)
    :type seed: int

    """

    def __init__(self, df, id_col, verbose=False, seed=None):

        self.df = df
        self.id_col = id_col
        self.seed = seed
        self.verbose = verbose

        if not self.seed:
            self.seed = int(round(1000 * np.random.random()))

    def extract(self, sample_size, sampling_strategy="random", stratify_by=None):

        """
        Extract a sample from a :py:class:`pandas.DataFrame` using one of the following methods:

        - all: Returns all of the IDs
        - random: Returns a random sample
        - stratify: Proportional stratification, method from Kish, Leslie. "Survey sampling." (1965). Chapter 4.
        - stratify_even: Sample evenly from each strata (will obviously not be representative)
        - stratify_guaranteed: Proportional stratification, but the sample is guaranteed to contain at least one \
            observation from each strata (if sample size is small and/or there are many small strata, the resulting \
            sample may be far from representative)

        :param sample_size: The desired size of the sample
        :type sample_size: int
        :param sampling_strategy: The method to be used to extract samples. Options are: all, random, stratify, \
        stratify_even, stratify_guaranteed
        :type sampling_strategy: str
        :param stratify_by: Optional name of a column or list of columns in the :py:class:`pandas.DataFrame` to \
        stratify on
        :type stratify_by: str, list
        :return: A list of IDs reflecting the observations selected from the :py:class:`pandas.DataFrame` during \
        sampling
        :rtype: list

        Usage::

            from pewanalytics.stats.sampling import SampleExtractor
            import nltk
            import pandas as pd

            nltk.download("inaugural")
            frame = pd.DataFrame([
                {"speech": fileid, "text": nltk.corpus.inaugural.raw(fileid)} for fileid in nltk.corpus.inaugural.fileids()
            ])
            frame["century"] = frame['speech'].map(lambda x: "{}00".format(x.split("-")[0][:2]))

            >>> sampler = SampleExtractor(frame, "speech", seed=42)

            >>> sample_index = sampler.extract(12, sampling_strategy="random")
            frame[frame["speech"].isin(sample_index)]['century'].value_counts()
            1900    6
            1800    5
            1700    1
            Name: century, dtype: int64

            >>> sample_index = sampler.extract(12, sampling_strategy="stratify", stratify_by=['century'])
            frame[frame["speech"].isin(sample_index)]['century'].value_counts()
            1800    5
            1900    5
            2000    1
            1700    1
            Name: century, dtype: int64

            >>> sample_index = sampler.extract(12, sampling_strategy="stratify_even", stratify_by=['century'])
            frame[frame["speech"].isin(sample_index)]['century'].value_counts()
            1800    3
            2000    3
            1700    3
            1900    3
            Name: century, dtype: int64

            >>> sample_index = sampler.extract(12, sampling_strategy="stratify_guaranteed", stratify_by=['century'])
            frame[frame["speech"].isin(sample_index)]['century'].value_counts()
            1900    5
            1800    4
            1700    2
            2000    1
            Name: century, dtype: int64

        """

        strategies = [
            "all",
            "random",
            "stratify",
            "stratify_even",
            "stratify_guaranteed",
        ]
        if sampling_strategy not in strategies:
            raise Exception(
                "You must choose one of the following sampling strategies: {}".format(
                    strategies
                )
            )

        doc_ids = None

        if sampling_strategy == "all":
            doc_ids = self.df[self.id_col].values

        elif sampling_strategy == "random":
            doc_ids = self._random_sample(sample_size).values

        elif sampling_strategy.startswith("stratify"):

            if self.verbose:
                print("Stratify on columns: {}".format(",".join(stratify_by)))

            self.df["_stratify_by"] = (
                self.df[stratify_by].astype(str).apply("".join, axis=1)
            )

            # So you can pass in a decimal proportion of total dataframe or number of samples
            sample_n = (
                sample_size
                if sample_size >= 1
                else int(round(sample_size * self.df.shape[0]))
            )

            if sampling_strategy == "stratify":
                doc_ids = self._stratify_sample(sample_n)

            elif sampling_strategy == "stratify_even":
                doc_ids = self._stratify_even_sample(sample_n)

            elif sampling_strategy == "stratify_guaranteed":
                doc_ids = self._stratify_guaranteed_sample(sample_n)

            del self.df["_stratify_by"]

        if self.verbose:
            print("Sample of %i extracted" % (len(doc_ids)))

        return list(doc_ids)

    def _random_sample(self, sample_size):

        if self.verbose:
            print("Basic random sample")
        if sample_size >= 1:
            return self.df.sample(int(sample_size), random_state=self.seed)[self.id_col]
        else:
            return self.df.sample(frac=float(sample_size), random_state=self.seed)[
                self.id_col
            ]

    def _stratify_sample(self, sample_size):

        if self.verbose:
            print("Kish-style stratification")

        # Subset & copy cols that we care about
        data = self.df.copy()[[self.id_col] + ["_stratify_by"]]
        frame_size = data.shape[0]

        # Shuffle the dataframe
        if self.verbose:
            print("Random seed: {}".format(self.seed))
        np.random.seed(self.seed)
        if self.verbose:
            print("Dataframe before sorting: {}".format(data.head()))
        data.index = np.random.permutation(data.index)

        # Re-sort grouped by strata
        data = data.groupby("_stratify_by").apply(lambda x: x.sort_index())
        data.index = list(range(0, frame_size))
        if self.verbose:
            print("Dataframe after shuffle & groupby sorting: {}".format(data.head()))

        skip_interval = float(frame_size) / float(sample_size)

        start_index = np.random.uniform(0, skip_interval)  # index to start from
        if self.verbose:
            print("Start index: {}".format(start_index))
        sample_index = np.round(
            (np.zeros(sample_size) + start_index)
            + (np.arange(sample_size) * skip_interval)
        )

        # Return the real id column
        sample_ids = data[data.index.isin(sample_index)][self.id_col].values

        return sample_ids

    def _stratify_even_sample(self, sample_size):

        random.seed(self.seed)
        docs_per_strata = int(
            float(sample_size)
            / float(self.df.groupby("_stratify_by")[self.id_col].count().count())
        )
        if self.verbose:
            print(
                "Drawing even samples of {} across all stratification groups".format(
                    docs_per_strata
                )
            )
        doc_ids = []
        for strata in self.df["_stratify_by"].unique():
            strata_data = self.df[self.df["_stratify_by"] == strata]
            doc_ids.extend(
                list(strata_data.sample(docs_per_strata)[self.id_col].values)
            )
        if len(doc_ids) < sample_size:
            doc_ids.extend(
                list(self.df.sample(sample_size - len(doc_ids))[self.id_col].values)
            )

        return doc_ids

    def _stratify_guaranteed_sample(self, sample_size):

        # Number of groups to stratify by must be less than the total sample size
        strata_groups = self.df.groupby("_stratify_by")[self.id_col].count().count()
        if sample_size > strata_groups:
            if self.verbose:
                print(
                    "Sampling one document per strata first ({} strata total)".format(
                        strata_groups
                    )
                )
            strata_one = (
                self.df.groupby("_stratify_by")
                .apply(lambda x: x.sample(1, random_state=self.seed))[self.id_col]
                .values
            )
        else:
            if self.verbose:
                print(
                    "There are more strata groups ({}) than things to sample: {}".format(
                        strata_groups, sample_size
                    )
                )
            strata_one = (
                self.df.groupby("_stratify_by")
                .apply(lambda x: x.sample(1, random_state=self.seed))[self.id_col]
                .sample(sample_size)
                .values
            )

        left_to_sample = sample_size - len(strata_one)
        if left_to_sample > 0:
            doc_ids = SampleExtractor(
                self.df[~self.df[self.id_col].isin(strata_one)],
                self.id_col,
                seed=self.seed,
                verbose=self.verbose,
            )._stratify_sample(left_to_sample)
            doc_ids = list(doc_ids) + list(strata_one)
        else:
            if self.verbose:
                print("Nothing left to sample, no stratification applied")
            doc_ids = strata_one

        return doc_ids



from __future__ import absolute_import

import re
import copy
import pandas as pd
import numpy as np
import scipy.sparse as sp

import nltk

from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import ENGLISH_STOP_WORDS
from nltk.corpus import wordnet

from tqdm import tqdm
from stopit import ThreadingTimeout as Timeout

try:
    from rapidfuzz import fuzz
except ImportError:
    from fuzzywuzzy import fuzz
from difflib import SequenceMatcher
from stopit import TimeoutException

from pewtils import is_null, is_not_null
from pewtils.http import strip_html
from pewtils import decode_text as _decode_text
from pewtils.regex import URL_REGEX

from pewanalytics.stats.clustering import (
    compute_hdbscan_clusters,
    compute_kmeans_clusters,
)
from pewanalytics.stats.mutual_info import compute_mutual_info
from pewanalytics.stats.dimensionality_reduction import get_lsa, get_pca


def has_fragment(text, fragment):

    """
    Checks whether a substring ("fragment") is contained within a larger string ("text"). Uses the \
    :py:func:`pewtils.decode_text` function to process both the text and the fragment when running this check.

    :param text: The text to search
    :type text: str
    :param fragment: The fragment to search for
    :type fragment: str
    :return: Whether or not the text contains the fragment
    :rtype: bool

    Usage::

        from pewanalytics.text import has_fragment

        text = "testing one two three"

        >>> has_fragment(text, "one two")
        True

        >>> has_fragment(text, "four")
        False
    """

    return any([(fragment in text), (_decode_text(fragment) in _decode_text(text))])


def remove_fragments(text, fragments, throw_loud_fail=False):

    """
    Iteratively remove fragments from a string.

    :param text: The text toremove the fragments from
    :type text: str
    :param fragments: A list of string fragments to search for and remove
    :type fragments: list
    :param throw_loud_fail: bool; whether or not to raise an error if text decoding fails (default=False)
    :type throw_loud_fail: bool
    :return: The original string, minus any parts that matched the fragments provided
    :rtype: str

    Usage::

        from pewanalytics.text import remove_fragments

        text = "testing one two three"

        >>> remove_fragments(text, ["one two"])
        "testing  three"

        >>> remove_fragments(text, ["testing", "three"])
        " one two "
    """

    for f in fragments:
        new_text = text.replace(f, "")
        # if the new text is the same as previous, try decoding
        if new_text == text:
            new_text = _decode_text(text, throw_loud_fail).replace(
                _decode_text(f, throw_loud_fail), ""
            )
        # if the new text is still the same as previous, then new text is None
        if new_text == text:
            new_text = None
        if new_text:
            text = new_text
    return text


def filter_parts_of_speech(text, filter_pos=None, exclude=False):

    """
    Retain words associated with parts of speech in the text if ``exclude=False``.
    If ``exclude=True``, exclude words associated with parts of speech.
    Default is Noun (NN), Proper Noun (NNP) and Adjective (JJ)

    | The full list of POS is here: https://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html

    :param text: The string to process
    :type text: str
    :param filter_pos: Array of part of speech tags (default is 'NN', 'NNP', and 'JJ')
    :type filter_pos: list
    :param exclude: If ``True``, the function will remove words that match to the specified parts of speech; by default \
    this function *filters to* POS matches instead.
    :return: A string comprised solely of words that matched (or did not match) to the specified parts of speech, \
    depending on the value of ``exclude``
    :rtype: str

    Usage::

        from pewanalytics.text import filter_parts_of_speech

        text = "This is a very exciting sentence that can serve as a functional example"

        >>> filter_parts_of_speech(text, filter_pos=["NN"])
        'sentence example'

        >>> filter_parts_of_speech(text, filter_pos=["JJ"], exclude=True)
        'This is a very sentence that can serve as a example'

    """

    if not filter_pos:
        filter_pos = ("NN", "NNP", "JJ")
    text = text.split()
    tagged_words = nltk.pos_tag(text)
    if not exclude:
        valid = [word[0] for word in tagged_words if word[1] in filter_pos]
    else:
        valid = [word[0] for word in tagged_words if word[1] not in filter_pos]
    return " ".join(valid)


def get_fuzzy_ratio(text1, text2, throw_loud_fail=False):

    """
    Uses Levenshtein Distance to calculate similarity of two strings.  Measures how the edit distance compares
    to the overall length of the texts. Uses the :py:mod:`fuzzywuzzy` library in Python 2, and the :py:mod:`rapidfuzz` \
    library in Python 3.

    :param text1: First string
    :type text1: str
    :param text2: Second string
    :type text1: str
    :param throw_loud_fail: bool; whether or not to raise an error if text decoding fails (default=False)
    :type throw_loud_fail: bool
    :return: The Levenshtein ratio between the two strings
    :rtype: float

    Usage::

        from pewanalytics.text import get_fuzzy_ratio

        text1 = "This is a sentence."
        text2 = "This is a slightly difference sentence."

        >>> get_fuzzy_ratio(text1, text2)
        64.28571428571428

    """

    try:
        return fuzz.ratio(text1, text2)
    except (UnicodeDecodeError, UnicodeEncodeError):
        return fuzz.ratio(
            _decode_text(text1, throw_loud_fail), _decode_text(text2, throw_loud_fail)
        )


def get_fuzzy_partial_ratio(text1, text2, throw_loud_fail=False, timeout=5):

    """
    Useful to calculate similarity of two strings that are of noticeably different lengths.  Allows for the possibility
    that one text is a subset of the other; finds the largest overlap and computes the Levenshtein ratio on that.

    :param text1: First string
    :type text1: str
    :param text2: Second string
    :type text2: str
    :param timeout: The number of seconds to wait before giving up
    :type timeout: int
    :param throw_loud_fail: bool; whether or not to raise an error if text decoding fails (default=False)
    :type throw_loud_fail: bool
    :return: The partial Levenshtein ratio between the two texts
    :rtype: float

    :accepts kwarg timeout:

    Usage::

        from pewanalytics.text import get_partial_fuzzy_ratio

        text1 = "This is a sentence."
        text2 = "This is a sentence, but with more text."

        >>> get_partial_fuzzy_ratio(text1, text2)
        100.0

    """

    partial_ratio = None
    with Timeout(timeout, swallow_exc=True):
        try:
            partial_ratio = fuzz.partial_ratio(text1, text2)
        except (UnicodeDecodeError, UnicodeEncodeError):
            partial_ratio = fuzz.partial_ratio(
                _decode_text(text1, throw_loud_fail),
                _decode_text(text2, throw_loud_fail),
            )
        return partial_ratio


class SentenceTokenizer(object):

    """
    Initializes a tokenizer that can be be used to break text into tokens using the ``tokenize`` function

    :param base_tokenizer: The tokenizer to use (default = NLTK's English Punkt tokenizer)
    :param regex_split_trailing: A compiled regex object used to define the end of sentences
    :param regex_split_leading: A compiled regex object used to define the beginning of sentences

    Usage::

        from pewanalytics.text import SentenceTokenizer
        import re

        text = "This is a sentence. This is another sentence - and maybe a third sentence. And yet a fourth sentence."

        >>> tokenizer = SentenceTokenizer()
        >>> tokenizer.tokenize(text)
        ['This is a sentence.',
         'This is another sentence - and maybe a third sentence.',
         'And yet a fourth sentence.']

        >>> tokenizer = SentenceTokenizer(regex_split_leading=re.compile(r"\-"))
        >>> tokenizer.tokenize(text)
        ['This is a sentence.',
         'This is another sentence',
         'and maybe a third sentence.',
         'And yet a fourth sentence.']

    """

    def __init__(
        self, base_tokenizer=None, regex_split_trailing=None, regex_split_leading=None
    ):
        self.base_tokenizer = (
            base_tokenizer
            if base_tokenizer
            else nltk.data.load("tokenizers/punkt/english.pickle")
        )
        self.regex_split_trailing = regex_split_trailing
        self.regex_split_leading = regex_split_leading

    def tokenize(self, text, throw_loud_fail=False, min_length=None):

        """
        Tokenizes the text.

        :param text: The text to tokenize
        :type text: str
        :param throw_loud_fail: Whether or not to raise an error if text decoding fails (default=False)
        :type throw_loud_fail: bool
        :param min_length: The minimum acceptable length of a sentence (if a token is shorter than this, it will be \
        considered part of the preceding sentence) (default=None)
        :type min_length: int
        :return: A list of sentences
        :rtype: list
        """

        text = _decode_text(text, throw_loud_fail)

        partial_tokens = []
        token_group = []
        for t in self.base_tokenizer.tokenize(text):
            if not self.regex_split_leading:
                partial_tokens.append(t)
            else:
                leaders = self.regex_split_leading.findall(t)
                token_group = []
                for subt_lead in self.regex_split_leading.split(t):
                    if subt_lead != "":
                        token_group.append(subt_lead)
                    if len(leaders) == 0 or subt_lead not in leaders:
                        partial_tokens.append("".join(token_group))
                        token_group = []
                if len(token_group) > 0:
                    partial_tokens.append("".join([t for t in token_group if t != ""]))
        if len(token_group) > 0:
            partial_tokens.append("".join([t for t in token_group if t != ""]))

        if not self.regex_split_trailing:
            final_tokens = partial_tokens
        else:
            final_tokens = []
            token_group = []
            for t in partial_tokens:
                trailers = self.regex_split_trailing.findall(t)
                token_group = []
                for subt_trail in self.regex_split_trailing.split(t):
                    if subt_trail != "":
                        token_group.append(subt_trail)
                    if len(trailers) == 0 or subt_trail in trailers:
                        final_tokens.append("".join(token_group))
                        token_group = []
                if len(token_group) > 0:
                    final_tokens.append("".join([t for t in token_group if t != ""]))
            if len(token_group) > 0:
                final_tokens.append("".join([t for t in token_group if t != ""]))

        final_tokens = [t.strip() for t in final_tokens]

        if min_length:
            final_tokens = [f for f in final_tokens if len(f) >= min_length]

        return final_tokens


class TextOverlapExtractor(object):

    """
    A helper class designed to identify overlapping sections between two strings.

    :param tokenizer: The tokenizer to use (default = SentenceTokenizer())
    """

    def __init__(self, tokenizer=None):

        if not tokenizer:
            self.tokenizer = SentenceTokenizer()
        else:
            self.tokenizer = tokenizer

    def get_text_overlaps(self, text1, text2, min_length=20, tokenize=True):

        """
        Extracts all overlapping segments of at least ``min_length`` characters between the two texts. If ``tokenize=True``
        then only tokens that appear fully in both texts will be extracted. For example:

        :param text1: A piece of text
        :type text1: str
        :param text2: Another piece of text to compare against the first
        :type text2: str
        :param min_length: The minimum size of the overlap to be considered (number of characters)
        :type min_length: int
        :param tokenize: If True, overlapping segments will only be included if they consist of atomic tokens; \
        overlaps that consist of only part of a token will be excluded. By default, the text is tokenize into \
        sentences based on punctuation. (default=True)
        :type tokenize: bool
        :return: A list of all of the identified overlapping segments
        :rtype: list

        Usage::

            from pewanalytics.text import TextOverlapExtractor

            text1 = "This is a sentence. This is another sentence. And a third sentence. And yet a fourth sentence."
            text2 = "This is a different sentence. This is another sentence. And a third sentence. But the fourth \
            sentence is different too."

            >>> extractor = TextOverlapExtractor()

            >>> extractor.get_text_overlaps(text1, text2, min_length=10, tokenize=False)
            [' sentence. This is another sentence. And a third sentence. ', ' fourth sentence']

            >>> extractor.get_text_overlaps(text1, text2, min_length=10, tokenize=True)
            ['This is another sentence.', 'And a third sentence.']

        """

        valid_tokens = None
        if tokenize:
            valid_tokens = [
                t.strip() for t in self.tokenizer.tokenize(". ".join([text1, text2]))
            ]
        fragments = []
        s = SequenceMatcher(None, text1, text2, autojunk=True)
        for block in s.get_matching_blocks():
            if block.size >= min_length:
                overlap = text1[block.a : (block.a + block.size)]
                if tokenize:
                    for token in self.tokenizer.tokenize(
                        overlap, min_length=min_length
                    ):
                        token = token.strip()
                        if not valid_tokens or token in valid_tokens:
                            fragments.append(token)
                elif len(overlap) >= min_length:
                    fragments.append(overlap)

        return fragments

    def get_largest_overlap(self, text1, text2):

        """
        Returns the largest overlapping segment of text between the two texts (this doesn't use the tokenizer).

        :param text1: A piece of text
        :type text1: str
        :param text2: Another piece of text to compare against the first
        :type text2: str
        :return: The largest substring that occurs in both texts
        :rtype: str

        Usage::

            from pewanalytics.text import TextOverlapExtractor

            text1 = "Overlaping section, unique text another overlapping section"
            text2 = "Overlapping section, another overlapping section"


            >>> extractor = TextOverlapExtractor()

            >>> extractor.get_largest_overlap(text1, text2)
            ' another overlapping section'

        """

        s = SequenceMatcher(None, text1, text2)
        pos_a, pos_b, size = s.find_longest_match(0, len(text1), 0, len(text2))
        return text1[pos_a : pos_a + size]


class TextCleaner(object):

    """
    A class for cleaning text up, in preparation for NLP, etc.  Attempts to decode the text.

    This function performs for the following cleaning tasks, in sequence:

        - Removes HTML tags (optional)
        - Decodes the text
        - Filters out specified parts of speech (optional)
        - Converts text to lowercase (optional)
        - Removes URLs (optional)
        - Expands contractions
        - Removes stopwords
        - Lemmatizes or stems (optional)
        - Removes words less than three characters
        - Removes punctuation
        - Consolidates whitespace

    :param process_method: Options are "lemmatize", "stem", or None (default = "lemmatize")
    :type process_method: str
    :param processor: A lemmatizer or stemmer with a "lemmatize" or "stem" function (default for \
    process_method="lemmatize" is nltk.WordNetLemmatizer(); default for process_method="stem" is nltk.SnowballStemmer())
    :param filter_pos: A list of WordNet parts-of-speech tags to keep; \
    if provided, all other words will be removed (default = None)
    :type filter_pos: list
    :param lowercase: Whether or not to lowercase the string (default = True)
    :type lowercase: bool
    :param remove_urls: Whether or not to remove URLs and links from the text (default = True)
    :type remove_urls: bool
    :param replacers: A list of tuples, each with a regex pattern followed by the string/pattern to replace them with. \
    Anything passed here will be used in addition to a set of built-in replacement patterns for common contractions.
    :param stopwords: The set of stopwords to remove (default = nltk.corpus.stopwords.words('english') combined with \
    sklearn.feature_extraction.stop_words.ENGLISH_STOP_WORDS). If an empty list is passed, no stopwords will be used.
    :type stopwords: set
    :param strip_html: Whether or not to remove contents wrapped in HTML tags (default = False)
    :type strip_html: bool
    :param tokenizer: Tokenizer to use (default = nltk.WhitespaceTokenizer())
    :type replacers: list
    :param throw_loud_fail: bool; whether or not to raise an error if text decoding fails (default=False)
    :type throw_loud_fail: bool

    Usage::

        from pewanalytics.text import TextCleaner

        text = "<body> \
            Here's some example text.</br>It isn't a great example, but it'll do. \
            Of course, there are plenty of other examples we could use though. \
            http://example.com \
            </body>"

        >>> cleaner = TextCleaner(process_method="stem")
        >>> cleaner.clean(text)
        'exampl is_not great exampl cours plenti exampl could use though'

        >>> cleaner = TextCleaner(process_method="stem", stopwords=["my_custom_stopword"], strip_html=True)
        >>> cleaner.clean(text)
        'here some exampl is_not great exampl but will cours there are plenti other exampl could use though'

        >>> cleaner = TextCleaner(process_method="lemmatize", strip_html=True)
        >>> cleaner.clean(text)
        'example is_not great example course plenty example could use though'

        >>> cleaner = TextCleaner(process_method="lemmatize", remove_urls=False, strip_html=True)
        >>> cleaner.clean(text)
        'example text is_not great example course plenty example could use though http example com'

        >>> cleaner = TextCleaner(process_method="stem", strip_html=False)
        >>> cleaner.clean(text)
        'example text is_not great example course plenty example could use though http example com'

        >>> cleaner = TextCleaner(process_method="stem", filter_pos=["JJ"], strip_html=True)
        >>> cleaner.clean(text)
        'great though'

    """

    def __init__(
        self,
        process_method="lemmatize",
        processor=None,
        filter_pos=None,
        lowercase=True,
        remove_urls=True,
        replacers=None,
        stopwords=None,
        strip_html=False,
        tokenizer=None,
        throw_loud_fail=False,
    ):

        self.tokenizer = tokenizer if tokenizer else nltk.WhitespaceTokenizer()
        self.replacers = replacers if replacers else []
        self.replacers.extend(
            [
                (r"won\'t", "will_not"),
                (r"can\'t", "cannot"),
                (r"i\'m", "i am"),
                (r"ain\'t", "is not"),
                (r"(\w+)\'ll", r"\g<1> will"),
                (r"(\w+)n\'t", r"\g<1>_not"),
                (r"(\w+)\'ve", r"\g<1> have"),
                (r"(\w+)\'re", r"\g<1> are"),
                (r"(\w+)\'d", r"\g<1> would"),
                (r"it\'s", "it is"),
            ]
        )  # Borrowed from NLTK cookbook
        self.replacers = [
            (re.compile(r"\b{}\b".format(regex[0])), regex[1])
            for regex in self.replacers
        ]
        if process_method == "lemmatize":
            self.processor = processor if processor else nltk.WordNetLemmatizer()
            self.process_func = self.processor.lemmatize
        elif process_method == "stem":
            self.processor = processor if processor else nltk.SnowballStemmer("english")
            self.process_func = self.processor.stem
        else:
            self.processor = None
            self.process_func = None
        if is_null(stopwords):
            stopwords = set.union(
                set(nltk.corpus.stopwords.words("english")), set(ENGLISH_STOP_WORDS)
            )
        self.stopword_regex = re.compile(
            r"\b({})\b".format(r"|".join([re.escape(s) for s in stopwords if s])),
            re.IGNORECASE,
        )
        if remove_urls:
            self.url_regex = URL_REGEX
        else:
            self.url_regex = None

        self.filter_pos = filter_pos
        self.lowercase = lowercase
        self.throw_loud_fail = throw_loud_fail
        self.strip_html = strip_html
        self.final_regex = re.compile(r"\w*\d\w*")

    def clean(self, text):
        """
        Cleans the text.

        :param text: The string to clean
        :type text: str
        :return: The cleaned string
        :rtype: str
        """

        # try to remove any html tags in the string
        if self.strip_html:
            text = strip_html(text)

        # try to encode everything as utf-8
        text = _decode_text(text, self.throw_loud_fail)

        if self.filter_pos:
            text = filter_parts_of_speech(text, self.filter_pos)
        if self.lowercase:
            text = str(text).lower()
        if self.url_regex:
            text = self.url_regex.sub(" ", text)

        for regex, replace in self.replacers:
            text = regex.sub(replace, text)  # expand contractions
        text = self.stopword_regex.sub("", text)
        text = re.sub(r"\W+", " ", text)  # remove punctuation
        text = self.tokenizer.tokenize(text)  # split on whitespace

        if self.processor:
            text = [self.process_func(word) for word in text]

        text = " ".join([word for word in text if len(word) > 2])
        if self.processor:
            text = self.stopword_regex.sub("", text)
            text = re.sub(r"\W+", " ", text)

        text = self.final_regex.sub("", text)
        return text


class TextDataFrame(object):

    """
    This is a class full of functions for working with dataframes of documents. It contains utilities for identifying \
    potential duplicates, identifying recurring segments of text, computing metrics like mutual information, \
    extracting clusters of documents, and more.

    Given a :py:class:`pandas.DataFrame` and the name of the column that contains the text to be analyzed, the \
    TextDataFrame will automatically produce a TF-IDF sparse matrix representation of the text upon initialization. \
    All other parameters are passed along to the scikit-learn TfidfVectorizer.

    .. tip:: For more info on the parameters it excepts, refer to the official scikit-learn `TfidfVectorizer \
    documentation \
    <https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html>`_.

    :param df: A :py:class:`pandas.DataFrame` of documents.  Must contain a column with text.
    :param text_column: The name of the column in the :py:class:`pandas.DataFrame` that contains the text
    :type text_column: str
    :param vectorizer_kwargs: All remaining keyword arguments are passed to TfidfVectorizer

    Usage::

        from pewanalytics.text import TextDataFrame
        import pandas as pd
        import nltk

        nltk.download("inaugural")
        df = pd.DataFrame([
            {"speech": fileid, "text": nltk.corpus.inaugural.raw(fileid)} for fileid in nltk.corpus.inaugural.fileids()
        ])
        # Let's remove new line characters so we can print the output in the docstrings
        df['text'] = df['text'].str.replace("\\n", " ")

        # And now let's create some additional variables to group our data
        df['year'] = df['speech'].map(lambda x: int(x.split("-")[0]))
        df['21st_century'] = df['year'].map(lambda x: 1 if x >= 2000 else 0)

        # And we'll also create some artificial duplicates in the dataset
        df = df.append(df.tail(2)).reset_index()

        >>> tdf = TextDataFrame(df, "text", stop_words="english", ngram_range=(1, 2))
        >>> tdf_dense = pd.DataFrame(tdf.tfidf.todense(), columns=tdf.vectorizer.get_feature_names()).head(5)
        >>> tdf_dense.loc[:, (tdf_dense != 0).any(axis=0)]
        	    14th	14th day	 abandon  abandon government... zeal inspires	zeal purity	zeal rely	zeal wisdom
        0	0.034014	0.034014	0.000000	       0.000000	...      0.000000	   0.000000	 0.000000	   0.000000
        1	0.000000	0.000000	0.000000	       0.000000	...      0.000000	   0.000000	 0.000000	   0.000000
        2	0.000000	0.000000	0.000000	       0.000000	...      0.000000	   0.000000	 0.000000	   0.000000
        3	0.000000	0.000000	0.020984	       0.030686	...      0.000000	   0.000000	 0.030686	   0.000000
        4	0.000000	0.000000	0.000000	       0.000000	...      0.026539	   0.026539	 0.000000	   0.026539

    """

    def __init__(self, df, text_column, **vectorizer_kwargs):
        self.corpus = df
        self.text_column = text_column
        self.vectorizer = TfidfVectorizer(decode_error="ignore", **vectorizer_kwargs)
        self.tfidf = self.vectorizer.fit_transform(df[text_column])

    def search_corpus(self, text):

        """
        Compares the provided text against the documents in the corpus and returns the most similar documents. \
        A new column called 'cosine_similarity' is generated, which is used to sort and return the \
        :py:class:`pandas.DataFrame`.

        :param text: The text to compare documents against
        :type text: str
        :return: The corpus :py:class:`pandas.DataFrame` sorted by cosine similarity

        Usage::

            >>> tdf.search_corpus('upright zeal')[:5]
                                                            text	search_cosine_similarity
            4	Proceeding, fellow citizens, to that qualifica...	0.030856
            8	Fellow citizens, I shall not attempt to descri...	0.025041
            9	In compliance with an usage coeval with the ex...	0.024922
            27	Fellow citizens, In obedience to the will of t...	0.021272
            10	Fellow citizens, about to undertake the arduou...	0.014791

        """

        similarities = cosine_similarity(self.vectorizer.transform([text]), self.tfidf)
        corpus = copy.deepcopy(self.corpus[[self.text_column]])
        corpus["search_cosine_similarity"] = similarities[0]
        return corpus.sort_values("search_cosine_similarity", ascending=False)

    def match_text_to_corpus(
        self, match_list, allow_multiple=False, min_similarity=0.9
    ):
        """
        Takes a list of text values and attempts to match them to the documents in the :py:class:`pandas.DataFrame`. \
        Each document will be matched to the value in the list to which it is most similar, based on cosine similarity.

        :param match_list: A list of strings (other documents) to be matched to documents in the \
        :py:class:`pandas.DataFrame`
        :type match_list: str
        :param allow_multiple: If set to True, each document in your corpus will be matched with its closes valid \
        match in the list. If set to False (default), documents in the list will only be matched to their best match \
        in the corpus.
        :type allow_multiple: bool
        :param min_similarity: Minimum cosine similarity required for any match to be made.
        :type min_similarity: float
        :return: Your corpus :py:class:`pandas.DataFrame`, with new columns match_text, match_index, \
        and cosine_similarity

        Usage::

            >>> match_df = tdf.match_text_to_corpus(test_excerpt, min_similarity=0.05)
            >>> match_df.sort_values('cosine_similarity')[:2]
                                                             text	                                       match_text	match_index	cosine_similarity
            48	Senator Hatfield, Mr. Chief Justice, Mr. Presi...	In this present crisis, government is not the ...	1	        0.0699283
            43	Vice President Johnson, Mr. Speaker, Mr. Chief...	And so, my fellow Americans: ask not what your...	0	        0.166681

        """
        similarities = cosine_similarity(
            self.tfidf, self.vectorizer.transform(match_list)
        )

        corpus = copy.deepcopy(self.corpus[[self.text_column]])
        corpus["match_text"] = None
        corpus["match_index"] = None
        corpus["cosine_similarity"] = None

        for index, row in tqdm(corpus.iterrows(), desc="Matching items to corpus"):
            row = corpus.iloc[index]
            if is_null(row["match_index"]):
                for i, sim in [
                    s
                    for s in sorted(
                        zip(
                            list(range(0, len(match_list) + 1)),
                            similarities[corpus.index.get_loc(index)],
                        ),
                        key=lambda x: x[1],
                        reverse=True,
                    )
                    if s[1] >= min_similarity
                ]:
                    match = True
                    if (
                        not allow_multiple
                        and i
                        in corpus[~corpus["match_index"].isnull()][
                            "match_index"
                        ].unique()
                    ):
                        current_best = corpus.loc[corpus["match_index"] == i][
                            "cosine_similarity"
                        ].max()
                        if sim >= current_best:
                            corpus.loc[corpus["match_index"] == i, "match_text"] = None
                            corpus.loc[
                                corpus["match_index"] == i, "cosine_similarity"
                            ] = None
                            corpus.loc[corpus["match_index"] == i, "match_index"] = None
                        else:
                            match = False
                    if match:
                        corpus.loc[
                            corpus[self.text_column] == row[self.text_column],
                            "match_index",
                        ] = i
                        corpus.loc[
                            corpus[self.text_column] == row[self.text_column],
                            "match_text",
                        ] = match_list[i]
                        corpus.loc[
                            corpus[self.text_column] == row[self.text_column],
                            "cosine_similarity",
                        ] = sim
                        break

        return corpus

    def extract_corpus_fragments(
        self,
        scan_top_n_matches_per_doc=20,
        min_fragment_length=15,
        tokenize=True,
        tokenizer=None,
    ):
        """
        Iterate over the corpus :py:class:`pandas.DataFrame` and, for each document, scan the most similar other \
        documents in the corpus using TF-IDF cosine similarity. During each comparison, overlapping fragments are \
        identified.  This can be useful for identifying common boilerplate sentences, repeated paragraphs, etc. \
        By default, the text is tokenized into complete sentences (so only complete sentences that recur will be \
        returned), but you can set ``tokenize=False`` to get raw segments of text that occur multiple times.

        :param scan_top_n_matches_per_doc: The number of other documents to compare each document against.
        :type scan_top_n_matches_per_doc: int
        :param min_fragment_length: The minimum character length a fragment must have to be extracted.
        :type min_fragment_length: int
        :param tokenize: If True, overlapping segments will only be included if they consist of atomic tokens; \
        overlaps that consist of only part of a token will be excluded. Uses sentence tokenization by default. \
        (default=True)
        :type tokenize: bool
        :param tokenizer: The tokenizer to use, if tokenizing isn't disabled (default = SentenceTokenizer())
        :type tokenizer: object
        :return: A list of fragments that were found.

        .. note:: This function will skip over duplicates if they exist in your data; it only compares documents
            that have less than .997 cosine similarity.

        Usage::

            >>> tdf.extract_corpus_fragments(scan_top_n_matches_per_doc=20, min_fragment_length=25, tokenize=False)
            ['s. Equal and exact justice ',
             'd by the General Government',
             ' of the American people, ',
             'ent of the United States ',
             ' the office of President of the United States ',
             ' preserve, protect, and defend the Constitution of the United States."  ',
             ' to "preserve, protect, and defend',
             ' of the United States are ',
             'e of my countrymen I am about to ',
             'Vice President, Mr. Chief Justice, ',
             ' 200th anniversary as a nation',
             ', and my fellow citizens: ',
             'e United States of America']
        """

        text_overlap_extractor = TextOverlapExtractor(tokenizer=tokenizer)

        similarity_matrix = cosine_similarity(self.tfidf)
        min_similarity = np.average([np.average(row) for row in similarity_matrix])

        combos = []
        for i in range(0, len(self.corpus.index)):
            combos.extend(
                [
                    (i, cos_similarity[0])
                    for cos_similarity in sorted(
                        zip(
                            list(range(i + 1, len(self.corpus.index))),
                            similarity_matrix[i][i + 1 :],
                        ),
                        reverse=True,
                    )
                    if min_similarity <= cos_similarity[1] < 0.997
                ][:scan_top_n_matches_per_doc]
            )
        fragments = []
        for i, cos_similarity in tqdm(combos, desc="Extracting fragments"):
            for frag in text_overlap_extractor.get_text_overlaps(
                self.corpus.iloc[i][self.text_column],
                self.corpus.iloc[cos_similarity][self.text_column],
                min_length=min_fragment_length,
                tokenize=tokenize,
            ):
                if frag not in fragments:
                    fragments.append(frag)

        return fragments

    def find_duplicates(
        self,
        tfidf_threshold=0.9,
        fuzzy_ratio_threshold=90,
        allow_partial=False,
        max_partial_difference=40,
        filter_function=None,
        partial_ratio_timeout=5,
        decode_text=False,
    ):

        """
        Search for duplicates by using cosine similarity and Levenshtein ratios.  This will struggle with large
        corpora, so we recommend trying to filter down to potential duplicates first.  The corpus will first be
        scanned for document pairs with a cosine similarity greater or equal to the ``tfidf_threshold``.  Then,
        each of these pairs will be compared using the more stringent ``fuzzy_ratio_threshold``.

        :param tfidf_threshold: Minimum cosine similarity for two documents to be considered potential dupes.
        :type tfidf_threshold: float
        :param fuzzy_ratio_threshold: The required Levenshtein ratio to consider two documents duplicates.
        :type fuzzy_ratio_threshold: int
        :param allow_partial: Whether or not to allow a partial ratio (if False, absolute ratios will be used)
        :type allow_partial: bool
        :param max_partial_diff: The maximum partial ratio difference allowed for a potential duplicate pair
        :type max_partial_diff: int
        :param filter_function: An optional function that allows for more complex filtering. The function must accept \
        the following parameters: text1, text2, cosine_similarity, fuzzy_ratio.  Must return True or False, \
        indicating whether the two documents should be considered duplicates.
        :param partial_ratio_timeout: How long, in seconds, that the partial ratio is allowed to compute
        :type partial_ratio_timeout: int
        :param decode_text: Whether to decode the text prior to making comparisons
        :type decode_text: bool
        :return: A list of lists, containing groups of duplicate documents (represented as rows from the corpus \
        :py:class:`pandas.DataFrame`)

        Usage::

            >>> tdf.find_duplicates()
            [           speech                                               text  year
            56  2013-Obama.txt  Thank you. Thank you so much.    Vice Presiden...  2013
            56  2013-Obama.txt  Thank you. Thank you so much.    Vice Presiden...  2013

                21st_century
            56             1
            56             1  ,
                        speech                                               text  year
            57  2017-Trump.txt  Chief Justice Roberts, President Carter, Presi...  2017
            57  2017-Trump.txt  Chief Justice Roberts, President Carter, Presi...  2017

                21st_century
            57             1
            57             1  ]

        """

        text = copy.deepcopy(self.corpus[self.text_column])
        if decode_text:
            text = text.map(_decode_text)

        groups = {}
        # compute cosine similarity between the inputs in tf.idf matrix
        similarity_matrix = cosine_similarity(self.tfidf)
        threshold_filter_matrix = similarity_matrix >= tfidf_threshold

        # return the  location of the similarity matrix that satisfies the threshold
        similarity_matrix = np.where(threshold_filter_matrix, similarity_matrix, None)

        # create pairs in the similarity matrix
        pairs = np.argwhere(similarity_matrix)
        pairs = sorted(pairs, key=lambda x: similarity_matrix[x[0]][x[1]], reverse=True)
        pairs = [p for p in pairs if p[0] > p[1]]

        for i, j in tqdm(pairs, desc="Scanning pairs"):
            sim = similarity_matrix[i][j]
            ratio = get_fuzzy_ratio(text.iloc[i], text.iloc[j])
            if ratio < fuzzy_ratio_threshold and allow_partial:
                try:
                    partial_ratio = get_fuzzy_partial_ratio(
                        text.iloc[i], text.iloc[j], timeout=partial_ratio_timeout
                    )
                except (MemoryError, TimeoutException):
                    partial_ratio = None
                except Exception as e:
                    print(e)
                    partial_ratio = None
                if (
                    partial_ratio
                    and abs(ratio - partial_ratio) <= max_partial_difference
                ):
                    ratio = max([ratio, partial_ratio])
            if ratio >= fuzzy_ratio_threshold and (
                not filter_function
                or filter_function(self.corpus.iloc[i], self.corpus.iloc[j], sim, ratio)
            ):
                if i not in list(groups.keys()) and j not in list(groups.keys()):
                    new_group = set([i, j])
                    groups[i] = new_group
                    groups[j] = new_group
                elif i in list(groups.keys()) and j not in list(groups.keys()):
                    groups[j] = groups[i]
                elif j in list(groups.keys()) and i not in list(groups.keys()):
                    groups[i] = groups[j]
                else:
                    groups[i].add(j)
                    groups[j].add(i)

        duplicates = []
        final_groups = []
        for g in groups.values():
            if g not in final_groups:
                final_groups.append(g)
                duplicates.append(self.corpus.iloc[list(g)])

        return duplicates

    def find_related_keywords(self, keyword, n=25):

        """
        Given a particular keyword, looks for related terms in the corpus using mutual information.

        :param keyword: The keyword to use
        :type keyword: str
        :param n: Number of related terms to return
        :type n: int
        :return: Terms associated with the keyword
        :rtype: list

        Usage::

            >>> tdf.find_related_keywords("war")[:2]
            ['war', 'peace']

            >>> tdf.find_related_keywords("economy")[:2]
            ['economy', 'expenditures']

        """

        self.corpus["temp"] = (
            self.corpus[self.text_column]
            .str.contains(r"\b{}\b".format(keyword), re.IGNORECASE)
            .astype(int)
        )
        mi = self.mutual_info("temp")
        del self.corpus["temp"]

        return list(mi[mi["MI1"] > 0].sort_values("MI1", ascending=False)[:n].index)

    def mutual_info(
        self, outcome_col, weight_col=None, sample_size=None, l=0, normalize=True
    ):

        """
        A wrapper around :py:func:`pewanalytics.stats.mutual_info.compute_mutual_info`

        :param outcome_col: The name of the column with the binary outcome variable
        :type outcome_col: str
        :param weight_col: (Optional) Name of the column to use in weighting
        :type weight_col: str
        :param sample_size: (Optional) If provided, a random sample of this size will be used instead of the full \
        :py:class:`pandas.DataFrame`
        :type sample_size: int
        :param l: An optional Laplace smoothing parameter
        :type l: float
        :param normalize: Toggle normalization on or off (to control for feature prevalence), on by default
        :type normalize: bool
        :return: A :py:class:`pandas.DataFrame` of ngrams and various metrics about them, including mutual information

        Usage::

            >>> results = tdf.mutual_info('21st_century')
            >>> results.sort_values("MI1", ascending=False).index[:25]
            Index(['journey complete', 'jobs', 'make america', 've', 'obama', 'workers',
                   'xand', 'states america', 'america best', 'debates', 'clinton',
                   'president clinton', 'trillions', 'stops right', 'transferring',
                   'president obama', 'stops', 'protected protected', 'transferring power',
                   'nation capital', 'american workers', 'politicians', 'people believe',
                   'borders', 'victories'],
                   dtype='object')

        """

        keep_columns = [self.text_column, outcome_col]
        if weight_col:
            keep_columns.append(weight_col)
        df = copy.deepcopy(self.corpus[keep_columns])
        if sample_size:
            df = df.sample(n=sample_size).reset_index()
        if weight_col:
            df = df.dropna().reset_index()
        else:
            df = df.dropna(subset=[self.text_column, outcome_col]).reset_index()
        y = df[outcome_col]
        x = self.vectorizer.transform(df[self.text_column])
        weights = None
        if weight_col:
            weights = df[weight_col]

        return compute_mutual_info(
            y,
            x,
            weights=weights,
            col_names=self.vectorizer.get_feature_names(),
            l=l,
            normalize=normalize,
        )

    def kmeans_clusters(self, k=10):

        """
        A wrapper around :py:func:`pewanalytics.stats.clustering.compute_kmeans_clusters`. Will compute clusters of documents.
        The resulting cluster IDs for each document are saved in the TextDataFrame's ``corpus`` in a new column called
        "kmeans".

        :param k: The number of clusters to extract
        :type k: int

        Usage::

            >>> tdf.kmeans_clusters(5)
            KMeans: n_clusters 5, score is 0.019735248210503934
            KMeans clusters saved to self.corpus['kmeans']

            >>> df['kmeans'].value_counts()
            2    26
            3    15
            4    11
            0     5
            1     3
            Name: kmeans, dtype: int64
        """

        self.corpus["kmeans"] = compute_kmeans_clusters(
            self.tfidf, k=k, return_score=False
        )
        print("KMeans clusters saved to self.corpus['kmeans']")

    def hdbscan_clusters(self, min_cluster_size=100, min_samples=1):

        """
        A wrapper around :py:func:`pewanalytics.stats.clustering.compute_hdbscan_clusters`. Will compute clusters \
        of documents. The resulting cluster IDs for each document are saved in the TextDataFrame's ``corpus`` in a \
        new column called "hdbscan".

        :param min_cluster_size: The minimum number of documents that a cluster must contain.
        :type min_cluster_size: int
        :param min_samples: An HDBSCAN parameter; refer to the documentation for more information
        :type min_samples: int

        Usage::

            >>> tdf.hdbscan_clusters(min_cluster_size=10)
            HDBSCAN: n_clusters 2
            HDBSCAN clusters saved to self.corpus['hdbscan']
        """

        self.corpus["hdbscan"] = compute_hdbscan_clusters(
            self.tfidf, min_cluster_size=min_cluster_size, min_samples=min_samples
        )
        print("HDBSCAN clusters saved to self.corpus['hdbscan']")

    def top_cluster_terms(self, cluster_col, min_size=50, top_n=10):

        """
        Extracts the top terms for each cluster, based on a column of cluster IDs saved to ``self.corpus``, using
        mutual information. Returns the ``top_n`` terms for each cluster.

        :param cluster_col: The name of the column that contains the document cluster IDs
        :type cluster_col: str
        :param min_size: Ignore clusters that have fewer than this number of documents
        :type min_size: int
        :param top_n: The number of top terms to identify for each cluster
        :type top_n: int
        :return: A dictionary; keys are the cluster IDs and values are the top terms for the cluster
        :rtype: dict

        Usage::

            >>> df_top_cluster = tdf.top_cluster_terms('kmeans', min_size=10)
            Cluster #2, 26 documents: ['constitution' 'union' 'states' 'friendly' 'liberal' 'revenue'
             'general government' 'confederacy' 'whilst' 'authorities']
            Cluster #4, 10 documents: ['shall strive' 'let sides' 'woe' 'offenses' 'breeze' 'war let'
             'nuclear weapons' 'learned live' 'mistakes' 'mr speaker']
            Cluster #0, 12 documents: ['activities' 'realization' 'interstate' 'wished' 'industrial' 'major'
             'counsel action' 'conditions' 'natural resources' 'eighteenth amendment']

        """

        dummies = pd.get_dummies(self.corpus[cluster_col], prefix=cluster_col)
        cluster_df = pd.concat([self.corpus, dummies], axis=1)

        terms = {}
        for cluster in cluster_df[cluster_col].unique():
            if (
                is_not_null(cluster)
                and len(cluster_df[cluster_df[cluster_col] == cluster]) >= min_size
            ):
                self.corpus["{}_{}".format(cluster_col, cluster)] = cluster_df[
                    "{}_{}".format(cluster_col, cluster)
                ]
                minfo = self.mutual_info("{}_{}".format(cluster_col, cluster))
                minfo = minfo.sort_values("MI1", ascending=False)[:top_n]
                del self.corpus["{}_{}".format(cluster_col, cluster)]
                minfo = minfo[minfo["MI1"] > 0].sort_values("MI1", ascending=False)[
                    :top_n
                ]
                terms[cluster] = minfo.index.values
                print(
                    "Cluster #{}, {} documents: {}".format(
                        cluster,
                        len(cluster_df[cluster_df[cluster_col] == cluster]),
                        minfo.index.values,
                    )
                )
        return terms

    def pca_components(self, k=20):

        """
        A wrapper around :py:func:`pewanalytics.stats.dimensionality_reduction.get_pca`.
        Saves the PCA components to self.corpus as new columns ('pca_1', 'pca_2', etc.),
        saves the top component for each document as self.corpus['pca'], and returns
        the features-component matrix.

        :param k: Number of dimensions to extract
        :type k: int
        :return: A :py:class:`pandas.DataFrame` of (features x components)

        Usage::

            >>> df_pca = tdf.pca_components(2)
            Decomposition explained variance ratio: 0.07488529151231405
            Component 0: ['america' 'today' 'americans' 'world' 'new' 'freedom' 'thank' 'nation'
             'god' 'journey']
            Component 1: ['america' 'make america' 'dreams' 'protected' 'obama' 'borders'
             'factories' 'american' 'transferring' 'stops']
            Top PCA dimensions saved as clusters to self.corpus['pca']

            >>> df.sample(5)
            	             speech	                                             text	year	21st_century	    pca_0      pca_1	  pca
            0	1789-Washington.txt	Fellow-Citizens of the Senate and of the House...	1789	0	            -0.129094	0.016984	pca_1
            21	1873-Grant.txt      Fellow-Citizens: Under Providence I have been ...	1873	0	            -0.097430	0.009559	pca_1
            49	1985-Reagan.txt  	Senator Mathias, Chief Justice Burger, Vice Pr...	1985	0	            0.163833	-0.020259	pca_0
            2	1797-Adams.txt    	When it was first perceived, in early times, t...	1797	0	            -0.140250	0.024844	pca_1
            20	1869-Grant.txt   	Citizens of the United States:    Your suffrag...	1869	0	            -0.114444	0.014419	pca_1
        """

        for col in self.corpus.columns:
            if col.startswith("pca_"):
                del self.corpus[col]
        components, documents = get_pca(
            self.tfidf, feature_names=self.vectorizer.get_feature_names(), k=k
        )
        for col in documents.columns:
            self.corpus[col] = documents[col]
        print("Top PCA dimensions saved as clusters to self.corpus['pca_'] columns")
        return components

    def lsa_components(self, k=20):

        """
        A wrapper around :py:func:`pewanalytics.stats.dimensionality_reduction.get_lsa`.
        Saves the LSA components to self.corpus as new columns ('lsa_1', 'lsa_2', etc.),
        saves the top component for each document as self.corpus['lsa'], and returns
        the features-component matrix

        :param k: Number of dimensions to extract
        :type k: int
        :return: A :py:class:`pandas.DataFrame` of (features x components)

        Usage::

            >>> df_lsa = tdf.lsa_components(2)
            Decomposition explained variance ratio: 0.04722850124656694
            Top features:
            Component 0: ['government' 'people' 'america' 'states' 'world' 'nation' 'shall'
             'country' 'great' 'peace']
            Component 1: ['america' 'today' 'americans' 'world' 'new' 'freedom' 'thank' 'nation'
             'god' 'journey']
            Top LSA dimensions saved as clusters to self.corpus['lsa_'] columns

            >>> df.sample(5)
            	            speech                                                 text    year	21st_century	lsa_0	   lsa_1	  lsa
            37	1937-Roosevelt.txt    When four years ago we met to inaugurate a Pre...    1937	           0 0.293068	0.040802	lsa_0
            8	1821-Monroe.txt       Fellow citizens, I shall not attempt to descri...    1821	           0 0.348465	-0.212382	lsa_0
            7	1817-Monroe.txt       I should be destitute of feeling if I was not ...    1817	           0 0.369249	-0.237231	lsa_0
            26	1893-Cleveland.txt    My Fellow citizens, in obedience of the mandat...    1893	           0 0.275778	-0.128497	lsa_0
            59	2017-Trump.txt        Chief Justice Roberts, President Carter, Presi...    2017	           1 0.342111	0.511687	lsa_1

        """

        for col in self.corpus.columns:
            if col.startswith("lsa_"):
                del self.corpus[col]
        components, documents = get_lsa(
            self.tfidf, feature_names=self.vectorizer.get_feature_names(), k=k
        )
        for col in documents.columns:
            self.corpus[col] = documents[col]
        print("Top LSA dimensions saved as clusters to self.corpus['lsa_'] columns")
        return components

    def get_top_documents(self, component_prefix="cluster", top_n=5):

        """
        Use after running :py:func:`pewanalytics.text.TextDataFrame.get_pca_components` or \
        :py:func:`pewanalytics.text.TextDataFrame.get_lsa_components`. Returns the ``top_n`` documents with \
        the highest scores for each components.

        :param component_prefix: 'lsa' or 'pca' (you must first run get_pca_components or get_lsa_components)
        :type component_prefix: str
        :param top_n: Number of documents to return for each component
        :type top_n: int
        :return: A dictionary where keys are the component, and values are the text values for the component's \
        ``top_n`` documents
        :rtype: dict

        Usage::

            >>> df_lsa_topdoc = tdf.get_top_documents("lsa")
            >>> {key: len(value) for key, value in lsa_topdoc.items()}
            {'lsa_0': 5, 'lsa_1': 4}

            >>> lsa_topdoc['lsa_1'][0]
            'Chief Justice Roberts, President Carter, President Clinton, President Bush, President Obama, fellow \
            Americans, and people of the world: Thank you.  We, the citizens of America...'

        """

        top_docs = {}
        for col in [
            c
            for c in self.corpus.columns
            if c.startswith("{}_".format(component_prefix))
        ]:
            docs = self.corpus[self.corpus[component_prefix] == col].sort_values(
                col, ascending=False
            )[:top_n]
            top_docs[col] = docs[self.text_column].values
        return top_docs

    def make_word_cooccurrence_matrix(
        self, normalize=False, min_frequency=10, max_frequency=0.5
    ):

        """
        Use to produce word co-occurrence matrices. Based on a helpful StackOverflow post:
        https://stackoverflow.com/questions/35562789/how-do-i-calculate-a-word-word-co-occurrence-matrix-with-sklearn

        :param normalize: If True, will be normalized
        :type normalize: bool
        :param min_frequency: The minimum document frequency required for a term to be included
        :type min_frequency: int
        :param max_frequency: The maximum proportion of documents containing a term allowed to include the term
        :type max_frequency: int
        :return: A matrix of (terms x terms) whose values indicate the number of documents in which two terms co-occurred

        Usage::

            >>> wcm = tdf.make_word_cooccurrence_matrix(min_frequency=25, normalize=True)
            # Find the top cooccurring pair of words
            >>> wcm.stack().index[np.argmax(wcm.values)]
            ('protection', 'policy')

        """

        text = self.corpus[self.text_column]
        cv = CountVectorizer(
            ngram_range=(1, 1),
            stop_words="english",
            min_df=min_frequency,
            max_df=max_frequency,
        )
        mat = cv.fit_transform(text)
        mat[
            mat > 0
        ] = (
            1
        )  # this makes sure that we're counting number of documents words have in common \
        # and not weighting by the frequency of one of the words in a single document, which can lead to spurious links
        names = cv.get_feature_names()
        mat = mat.T * mat  # compute the document-document matrix
        if normalize:
            diag = sp.diags(1.0 / mat.diagonal())
            mat = diag * mat
        mat.setdiag(0)
        matrix = pd.DataFrame(data=mat.todense(), columns=names, index=names)

        return matrix

    def make_document_cooccurrence_matrix(self, normalize=False):

        """
        Use to produce document co-occurrence matrices. Based on a helpful StackOverflow post:
        https://stackoverflow.com/questions/35562789/how-do-i-calculate-a-word-word-co-occurrence-matrix-with-sklearn

        :param normalize: If True, will be normalized
        :type normalize: bool
        :return: A matrix of (documents x documents) whose values indicate the number of terms they had in common

        Usage::

            >>> dcm = tdf.make_document_cooccurrence_matrix(normalize=True)

            # Remove artifical duplicates and insert document names
            >>> dcm = dcm.iloc[:-2, :-2]
            >>> dcm.rename(columns=df['speech'][:-2],
                           index=df['speech'][:-2],
                           inplace=True)

            # Find documents with the highest coocurrence score
            >>> dcm.stack().index[np.argmax(dcm.values)]
            ('1793-Washington.txt', '1841-Harrison.txt')

        """

        text = self.corpus[self.text_column]
        cv = CountVectorizer(ngram_range=(1, 1), stop_words="english")
        mat = cv.fit_transform(text)
        mat[
            mat > 0
        ] = (
            1
        )  # this makes sure that we're counting number of words documents have in common \
        # and not weighting by the frequency of one of the words in a single document, which can lead to spurious links
        names = text.index
        mat = mat * mat.T  # compute the word-word matrix
        if normalize:
            diag = sp.diags(1.0 / mat.diagonal())
            mat = diag * mat
        mat.setdiag(0)
        matrix = pd.DataFrame(data=mat.todense(), columns=names, index=names)

        return matrix


def is_probable_stopword(word):
    """
    Determine if a word is likely to be a stop word (like a name of a person or location) by the following rules:

    1. Number of synset (words with similar meaning) is less than 3
    2. The min_depth (number of edges between a word and the top of the hierarchy) is > 5
    3. The number of lemma (similar to term definition in dictionary) is less than 2

    If more than one of these conditions is true, then this function will return False, because the word likely has
    one or more meanings in English and is likely to be more than just a proper name.

    This function was developed through trial and error, and your mileage may vary. It's intended to
    help you identify potential stopwords when extracting features from a database. For example, on one
    of our projects we wanted to remove names from our text data, and pulled a list of names from our
    database of politicians. However, some politicians have last names that are also common English words,
    like "White" and "Black" - and in those cases, we didn't want to add those to our list of stopwords.
    This function was useful in scanning through our list of names to identify names that we wanted to
    "whitelist".

    :param word: A word, usually a name of a person or location or something that you might want to add as a stopword
    :type word: string
    :return: Whether or not the word is (probably) a stopword aka a proper noun with no common English meaning
    :rtype: bool

    Usage::

        >>> is_probable_stopword("Chicago")
        True

        >>> is_probable_stopword("Chicago")
        False

        >>> is_probable_stopword("Orange")
        False

        >>> is_probable_stopword("Johnny")
        True

    """

    word = word.lower()
    synsets = wordnet.synsets(word)
    if not synsets or len(synsets) <= 1:
        return True
    else:
        total_synsets = len(synsets)
        min_depth = min([syn.min_depth() for syn in synsets])
        max_lemma_count = max(
            [sum([lemma.count() for lemma in syn.lemmas()]) for syn in synsets]
        )

        score = 0
        if total_synsets < 3:
            score += 1
        if min_depth >= 5:
            score += 1
        if max_lemma_count <= 2:
            score += 1

        if score > 1:
            return True

        else:
            return False
